#!/usr/bin/perl -w
# file packer/pack.pl
#
# Copyright 2009-2011 Dalton Harvie (daltonh@unimelb.edu.au)
# 
# This file is part of arb finite volume solver, referred to as `arb'.
# 
# arb is a software package designed to solve arbitrary partial
# differential equations on unstructured meshes using the finite volume
# method.  Primarily it consists of fortran source code, perl source
# code and shell scripts.  arb replies on certain third party software
# to run, most notably the computer algebra system maxima
# <http://maxima.sourceforge.net/> which is released under the GNU GPL.
# 
# The copyright of arb is held by Dalton Harvie.
# 
# arb is released under the GNU GPL.  arb is free software: you can
# redistribute it and/or modify it under the terms of the GNU General
# Public License (version 3) as published by the Free Software Foundation.
# You should have received a copy of the GNU General Public License
# along with arb (see file license/gpl.txt after unpacking).  If not,
# see <http://www.gnu.org/licenses/>.
# 
# arb is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
# FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
# 
# For full details of arb's license see the license directory.
# 
# The current homepage for the arb finite volume solver project is
# <http://www.chemeng.unimelb.edu.au/people/staff/daltonh/downloads/arb>.
#
#-------------------------------------------------------------------------
# perl script to pack up the bits of arb into one tar file ready for transport or achive
# daltonh, 300610

# requires: perl, rsync (>=2.6.7), tar, gzip

use strict;
#use File::Basename;
my ($archive_name, $version, $day, $month, $date, $argument, $directory, $archive_root, $systemcall);

print "pack perl script for packing components of arb into self-contained tar archive\n";

# set default options
my $tar = 1; # by default tar up files

# create a list of first level directories that must be created regardless of existing contents
my @directories=( 'src','doc','build','output','tmp','packer','examples','misc' );

# these are includes which are additional to the pack.include list
my $includes = "";

# unless distribute is on, the following free packages are also be included
$includes = $includes.' --include="src/" --include="src/contributed/" --include="src/contributed/suitesparse/" --include="src/contributed/suitesparse/*.tar.gz"';

# do sanity check that the directories exist, indicating that we are in the root directory
foreach $directory (@directories) {
  if (!(-d $directory)) {
    print "ERROR: the required component directory $directory is not found at this location.\n".
          "Are you sure you are calling pack from the simulation root directory (which contains the directory src for example)\n".
          "For more info on usage of pack try \"pack --help\"\n";
  }
}

my $include_file = "packer/pack.include";
if (!(-e $include_file)) { die "ERROR: include file $include_file not found\n"; }

# read through all command line arguments
foreach $argument ( @ARGV )  # loop through command line arguments
{
# print "$argument\n";
  if ( $argument eq '-h' || $argument eq '--help' ) {
    usage();
  } elsif ( $argument eq '-a' || $argument eq '--all') {
    print "INFO: all example, misc and linear solver files will be included in archive\n";
    $includes = $includes.' --include="examples/***" --include="misc/***" --include="src/" --include="src/contributed/***"';
  } elsif ( $argument eq '-d' || $argument eq '--distribute') {
    print "INFO: only distributable files will be included in archive\n";
    $includes = "";
  } elsif ( $argument eq '-b' || $argument eq '--build') {
    print "INFO: all build files will be included in archive\n";
    $includes = $includes.' --include="build/***"';
  } elsif ( $argument eq '-m' || $argument eq '--misc') {
    print "INFO: all misc files will be included in archive\n";
    $includes = $includes.' --include="misc/***"';
  } elsif ( $argument eq '-e' || $argument eq '--examples') {
    print "INFO: all example input files will be included in archive\n";
    $includes = $includes.' --include="examples/***"';
  } elsif ( $argument eq '-c' || $argument eq '--contributed') {
    print "INFO: all contributed files will be included in archive\n";
    $includes = $includes.' --include="src/" --include="src/contributed/***"';
  } elsif ( $argument eq '-t' || $argument eq '--notar' ) {
    print "INFO: archive will not be tarred\n";
    $tar = 0;
  } else {
    print "INFO: setting archive directory to $argument\n";
    $archive_name = $argument;
  }
}

# create archive_name using current version and date if not set from command line
if (!($archive_name)) {
  open(SRC, "<src/general_module.f90") or die "ERROR: problem opening src/main.f90 to find version number\n";
  while (<SRC>) {
    if (/version\s*=\s*(\S+)/) {
      $version = $1;
      last;
    }
  }
  if (!($version)) { die "ERROR: version number not found in src/general_module.f90\n"; }
  $day = scalar((localtime)[3]);
  if ($day < 10) {$day = "0".$day;}
  $month = scalar((localtime)[4]+1);
  if ($month < 10) {$month = "0".$month;}
# $date=$day.$month.scalar((localtime)[5]+1900);
  $date=scalar((localtime)[5]+1900).$month.$day;
  $archive_name = "arb_".$date."_".$version;
}
    
# remove archive_name that is empty
if (-d $archive_name) {rmdir $archive_name;}
# create new archive_name
if (-e $archive_name) {
  die "ERROR: $archive_name already exists: choose another name or delete this file/directory\n";
} else {
  mkdir $archive_name;
  print "INFO: created directory $archive_name\n";
  $archive_root = $archive_name.'/archive/';
  mkdir $archive_root;
}
  
# create required directories
foreach $directory (@directories) {
  mkdir $archive_root.$directory;
}
  
# now rsync over necessary files and subdirectories into archive_root
print "INFO: creating archive of files\n";
# basic files
$systemcall="rsync -auR --include-from=$include_file * $archive_root";  
(!(system($systemcall))) or die "ERROR: could not perform $systemcall\n";
# any extra files to include
if ($includes) {
  $systemcall="rsync -auR --exclude=\"recent\" --exclude=\"old\"$includes --exclude=\"*\" * $archive_root";  
  (!(system($systemcall))) or die "ERROR: could not perform $systemcall\n";
}

# record pack info to packer/history file within the archive
$systemcall='echo "PACK operation completed on `hostname -f` by `whoami` on `date` to archive name '.$archive_name.
  " with options @ARGV\" >>".$archive_root."packer/history";
(!(system($systemcall))) or die "ERROR: could not perform $systemcall\n";

if ($tar) {
  print "INFO: creating tar archive file\n";
  chdir $archive_root;
# $systemcall="tar -c -f ../archive.tar *"; # tar
  $systemcall="tar -cz -f ../archive.tar.gz *"; # tar and compress
  (!(system($systemcall))) or die "ERROR: could not perform $systemcall\n";
  chdir "..";
  $systemcall="rm -r archive";
  (!(system($systemcall))) or die "ERROR: could not perform $systemcall\n";
  chdir "..";
}

$systemcall="cp packer/unpack $archive_name";
(!(system($systemcall))) or die "ERROR: could not perform $systemcall\n";
$systemcall="cp packer/unpack_readme $archive_name/README";
(!(system($systemcall))) or die "ERROR: could not perform $systemcall\n";
$systemcall="cp license/arb_license.txt $archive_name/LICENSE";
(!(system($systemcall))) or die "ERROR: could not perform $systemcall\n";


print "INFO: success\n";

exit;

#*******************************************************************************
# usage function

sub usage {
  print "\nusage:\n\n".
        "pack [options] <archive_name>\n\n".
        "where <archive_name> is the name of the archive directory to be created,".
        "which must not already exist\n\n".
        "options include:\n".
        " -h or --help = produce this message\n".
        " -e or --examples = include all files within the examples directory\n".
        " -c or --contributed = include all files within the src/contributed directory\n".
        " -b or --build = include all files within the build directory\n".
        " -m or --misc = include all files within the misc directory\n".
        " -a or --all = equivalent to --examples --misc --contributed\n\n".
        " -d or --distribute = only distributable files will be included in the archive\n\n".
        "pack must be called from the root directory of a simulation\n";
  exit;
}

#*******************************************************************************
