! file src/contributed/intel_pardiso/intel_pardiso_module.f90
!
! Copyright 2009-2011 Dalton Harvie (daltonh@unimelb.edu.au)
! 
! This file is part of arb finite volume solver, referred to as `arb'.
! 
! arb is a software package designed to solve arbitrary partial
! differential equations on unstructured meshes using the finite volume
! method.  Primarily it consists of fortran source code, perl source
! code and shell scripts.  arb replies on certain third party software
! to run, most notably the computer algebra system maxima
! <http://maxima.sourceforge.net/> which is released under the GNU GPL.
! 
! The copyright of arb is held by Dalton Harvie.
! 
! arb is released under the GNU GPL.  arb is free software: you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License (version 3) as published by the Free Software Foundation.
! You should have received a copy of the GNU General Public License
! along with arb (see file license/gpl.txt after unpacking).  If not,
! see <http://www.gnu.org/licenses/>.
! 
! arb is distributed in the hope that it will be useful, but WITHOUT
! ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
! FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
! for more details.
! 
! For full details of arb's license see the license directory.
! 
! The current homepage for the arb finite volume solver project is
! <http://www.chemeng.unimelb.edu.au/people/staff/daltonh/downloads/arb>.
!
!-------------------------------------------------------------------------
module intel_pardiso_module

!-----------------------------------------------------------------
contains

!-----------------------------------------------------------------

subroutine intel_pardiso_linear_solver(a,ia,ja,x,error,ooc)

! this routine based on pardiso_unsym.f routine from pardiso documentation
!  to call pardiso solver packaged with the intel mkl libraries

implicit none
double precision, dimension(:), allocatable :: a, x
integer, dimension(:), allocatable :: ia, ja
integer :: n, nz, error, i
! variables required by pardiso
integer :: maxfct = 1 ! number of factoring matrices kept in memory
integer :: mnum = 1 ! particular matrix to factorise from those in memory
integer :: mtype = 11 ! type of matrix (unsymmetric)
integer :: msglvl = 0 ! how much information is printed (quiet=0)
integer :: nrhs = 1 ! number of b matrices
integer :: phase
integer*8, dimension(64) :: pt
integer, dimension(64) :: iparm
integer, dimension(:), allocatable :: perm
double precision, dimension(:), allocatable :: x_dummy
logical, optional :: ooc ! whether out of core solver is required
logical :: debug = .false., debug_sparse = .true.

external pardisoinit
external pardiso

if (debug) debug_sparse = .true.
if (debug_sparse) write(*,'(80(1h+)/a)') 'subroutine intel_pardiso_linear_solver'

n = ubound(ia,1) - 1
nz = ia(n+1) - 1
if (debug) write(*,*) 'n = ',n,': nz = ',nz
if (ubound(a,1) /= nz.or.ubound(ja,1) /= nz.or.ubound(x,1) /= n) stop &
  'ERROR: some of the csr matrices that are entering intel_pardiso_linear_solver are incorrectly sized'

allocate(perm(n),x_dummy(n))

! intitalise pointers (pt) and set most values to defaults (iparm)
do i=1,64
  pt(i) = 0
  iparm(i) = 0
end do
iparm(3) = 1 ! Numbers of Processors ( value of OMP_NUM_THREADS )
if (debug) msglvl = 1

if (debug_sparse) write(*,*) 'calling pardisoinit'

call pardisoinit(pt, mtype, iparm)

! additional options
! iparm(1) = 1 ! allows non-default options - already set from pardisoinit
if (debug) iparm(27) = 1 ! checks data structure
iparm(6) = 1 ! stores solution in x rather than x_dummy
if (present(ooc)) then
  if (ooc) iparm(60) = 2 ! use OOC pardiso version
end if

if (debug_sparse) write(*,*) 'done pardisoinit'
if (debug) then
  do i = 1, 64
    write(83,*) 'iparm(',i,') = ',iparm(i)
  end do
end if
    
! Reordering and Symbolic Factorization, This step also allocates
!  all memory that is necessary for the factorization 
phase = 11
call pardiso (pt, maxfct, mnum, mtype, phase, n, a, ia, ja, perm, nrhs, iparm, msglvl, x, x_dummy, error)
     
if (error /= 0) then
  write(*,*) 'ERROR: error in reordering and symbolic factorization of intel pardiso: error = ',error
  deallocate(perm,x)
  return
else
  if (debug_sparse) write(*,*) 'Reordering completed ... '
  if (debug) then
    write(*,*) 'Number of nonzeros in factors   = ',iparm(18)
    write(*,*) 'Number of factorization MFLOPS  = ',iparm(19)
  end if
end if

! Factorization
phase = 22
call pardiso (pt, maxfct, mnum, mtype, phase, n, a, ia, ja, perm, nrhs, iparm, msglvl, x, x_dummy, error)

if (error /= 0) then
  write(*,*) 'ERROR: error in factorization of intel pardiso: error = ',error
  deallocate(perm,x_dummy)
  return
else
  if (debug_sparse) write(*,*) 'Factorization completed ...  '
end if

! Back substitution and iterative refinement
phase = 33
! iparm(8)  = 1   ! max numbers of iterative refinement steps

call pardiso (pt, maxfct, mnum, mtype, phase, n, a, ia, ja, perm, nrhs, iparm, msglvl, x, x_dummy, error)
if (error /= 0) then
  write(*,*) 'ERROR: error in solve of intel pardiso: error = ',error
  deallocate(perm,x_dummy)
  return
else
  if (debug_sparse) write(*,*) 'Solve completed ... '
end if

! Termination and release of memory
phase = -1
call pardiso (pt, maxfct, mnum, mtype, phase, n, a, ia, ja, perm, nrhs, iparm, msglvl, x, x_dummy, error)

if (error /= 0) then
  write(*,*) 'ERROR: error in termination of intel pardiso: error = ',error
else
  if (debug_sparse) write(*,*) 'Termination completed ... '
end if

deallocate(perm,x_dummy)

if (debug_sparse) write(*,'(a/80(1h-))') 'subroutine intel_pardiso_linear_solver'

end subroutine intel_pardiso_linear_solver

!-----------------------------------------------------------------

end module intel_pardiso_module
!-----------------------------------------------------------------
