! file src/gmsh_module.f90
!
! Copyright 2009-2011 Dalton Harvie (daltonh@unimelb.edu.au)
! 
! This file is part of arb finite volume solver, referred to as `arb'.
! 
! arb is a software package designed to solve arbitrary partial
! differential equations on unstructured meshes using the finite volume
! method.  Primarily it consists of fortran source code, perl source
! code and shell scripts.  arb replies on certain third party software
! to run, most notably the computer algebra system maxima
! <http://maxima.sourceforge.net/> which is released under the GNU GPL.
! 
! The copyright of arb is held by Dalton Harvie.
! 
! arb is released under the GNU GPL.  arb is free software: you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License (version 3) as published by the Free Software Foundation.
! You should have received a copy of the GNU General Public License
! along with arb (see file license/gpl.txt after unpacking).  If not,
! see <http://www.gnu.org/licenses/>.
! 
! arb is distributed in the hope that it will be useful, but WITHOUT
! ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
! FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
! for more details.
! 
! For full details of arb's license see the license directory.
! 
! The current homepage for the arb finite volume solver project is
! <http://www.chemeng.unimelb.edu.au/people/staff/daltonh/downloads/arb>.
!
!-------------------------------------------------------------------------
module gmsh_module

! statements specifying different data types and parameters
implicit none

type node_list_type
  integer, dimension(:), allocatable :: node
end type node_list_type
  
! this data structure holds info about how elements, faces and nodes are defined in gmsh
type gtype_list_type
  character(len=100) :: name
  integer :: dimensions ! 0 for point, 1 for line, 2 for plane, 3 for volume
  integer :: nnodes ! number of surrounding nodes
  integer :: nfaces ! number of surrounding faces
  type(node_list_type), dimension(:), allocatable :: face_nodes ! lists the node indicies of each of the nfaces faces
  logical :: supported ! suitable for arb?
end type gtype_list_type

! reference list of all the gtypes
type(gtype_list_type), dimension(:), allocatable :: gtype_list

! info about each of the gelements
type gelement_type
  integer :: icell ! corresponding icell, 0 if none
  integer :: jface ! corresponding jface, 0 if none
  integer, dimension(:), allocatable :: gregions ! list of gregions that it is a member of, or failing that, one entry of 0
end type gelement_type

! type for each gmsh that is read in or output
type gmesh_type
  character(len=1000) :: filename
  character(len=1000) :: basename
  character(len=100), dimension(:), allocatable :: options ! array of options for this gmesh
  integer :: dimensions
  integer :: ngnodes ! number of gnodes
  integer, dimension(:), allocatable :: knode_from_gnode ! 0 if undefined
  integer, dimension(:), allocatable :: gnode_from_knode ! 0 if undefined
  integer :: ngelements ! number of gelements, including repeats for multiple regions
  integer :: ngelements_face ! number of face gelements, including repeats for multiple regions
  integer :: ngelements_cell ! number of cell gelements, including repeats for multiple regions
  type(gelement_type), dimension(:), allocatable :: gelement ! list of gelements
  integer :: ngregions ! number of gregions
  integer, dimension(:), allocatable :: region_from_gregion ! mapping from gregion number to arb region number, dimension gregion_max
end type gmesh_type

! gmsh meshes
type(gmesh_type), dimension(:), allocatable :: gmesh

! group lists of gmesh option types
character(len=100), dimension(:), allocatable :: output_gmesh_options, data_options, input_gmesh_options, output_options, &
  input_options

!-----------------------------------------------------------------
contains

!-----------------------------------------------------------------

subroutine setup_gtype_list

! here we allocate data to the reference gtype_list array

integer :: n
logical, parameter :: debug = .false.

if (debug) write(*,'(80(1h+)/a)') 'subroutine setup_gtype_list'

allocate(gtype_list(31))

! default is that element is unsupported
do n=1,ubound(gtype_list,1)
  gtype_list(n)%name = "unknown"
  gtype_list(n)%supported = .false.
end do

! now set supported elements
! faces are definined in the context of the element having the same dimensions as the problem
! ie, as a cell
! if the element has a dimension that is less than the problem dimension, it is assumed to be a
! face (dimension-1)
! surface nodes are directed using the right-hand rule with a positive outward pointing normal

n = 15
gtype_list(n)%name = "1-node point"
gtype_list(n)%dimensions = 0
gtype_list(n)%nnodes = 1
gtype_list(n)%nfaces = 0
gtype_list(n)%supported = .true.

n = 1
gtype_list(n)%name = "2-node line"
gtype_list(n)%dimensions = 1
gtype_list(n)%nnodes = 2
gtype_list(n)%nfaces = 2
allocate(gtype_list(n)%face_nodes(gtype_list(n)%nfaces))
allocate(gtype_list(n)%face_nodes(1)%node(1))
gtype_list(n)%face_nodes(1)%node = [ 1 ]
allocate(gtype_list(n)%face_nodes(2)%node(1))
gtype_list(n)%face_nodes(2)%node = [ 2 ]
gtype_list(n)%supported = .true.

n = 2
gtype_list(n)%name = "3-node triangle"
gtype_list(n)%dimensions = 2
gtype_list(n)%nnodes = 3
gtype_list(n)%nfaces = 3
allocate(gtype_list(n)%face_nodes(gtype_list(n)%nfaces))
allocate(gtype_list(n)%face_nodes(1)%node(2))
gtype_list(n)%face_nodes(1)%node = [ 1, 2 ]
allocate(gtype_list(n)%face_nodes(2)%node(2))
gtype_list(n)%face_nodes(2)%node = [ 2, 3 ]
allocate(gtype_list(n)%face_nodes(3)%node(2))
gtype_list(n)%face_nodes(3)%node = [ 3, 1 ]
gtype_list(n)%supported = .true.

n = 3
gtype_list(n)%name = "4-node quadrangle"
gtype_list(n)%dimensions = 2
gtype_list(n)%nnodes = 4
gtype_list(n)%nfaces = 4
allocate(gtype_list(n)%face_nodes(gtype_list(n)%nfaces))
allocate(gtype_list(n)%face_nodes(1)%node(2))
gtype_list(n)%face_nodes(1)%node = [ 1, 2 ]
allocate(gtype_list(n)%face_nodes(2)%node(2))
gtype_list(n)%face_nodes(2)%node = [ 2, 3 ]
allocate(gtype_list(n)%face_nodes(3)%node(2))
gtype_list(n)%face_nodes(3)%node = [ 3, 4 ]
allocate(gtype_list(n)%face_nodes(4)%node(2))
gtype_list(n)%face_nodes(4)%node = [ 4, 1 ]
gtype_list(n)%supported = .true.

n = 4
gtype_list(n)%name = "4-node tetrahedron"
gtype_list(n)%dimensions = 3
gtype_list(n)%nnodes = 4
gtype_list(n)%nfaces = 4
allocate(gtype_list(n)%face_nodes(gtype_list(n)%nfaces))
allocate(gtype_list(n)%face_nodes(1)%node(3))
gtype_list(n)%face_nodes(1)%node = [ 1, 4, 3 ]
allocate(gtype_list(n)%face_nodes(2)%node(3))
gtype_list(n)%face_nodes(2)%node = [ 2, 3, 4 ]
allocate(gtype_list(n)%face_nodes(3)%node(3))
gtype_list(n)%face_nodes(3)%node = [ 1, 2, 4 ]
allocate(gtype_list(n)%face_nodes(4)%node(3))
gtype_list(n)%face_nodes(4)%node = [ 1, 3, 2 ]
gtype_list(n)%supported = .true.

n = 5
gtype_list(n)%name = "8-node hexahedron"
gtype_list(n)%dimensions = 3
gtype_list(n)%nnodes = 8
gtype_list(n)%nfaces = 6
allocate(gtype_list(n)%face_nodes(gtype_list(n)%nfaces))
allocate(gtype_list(n)%face_nodes(1)%node(4))
gtype_list(n)%face_nodes(1)%node = [ 1, 5, 8, 4 ]
allocate(gtype_list(n)%face_nodes(2)%node(4))
gtype_list(n)%face_nodes(2)%node = [ 5, 6, 7, 8 ]
allocate(gtype_list(n)%face_nodes(3)%node(4))
gtype_list(n)%face_nodes(3)%node = [ 2, 3, 7, 6 ]
allocate(gtype_list(n)%face_nodes(4)%node(4))
gtype_list(n)%face_nodes(4)%node = [ 1, 4, 3, 2 ]
allocate(gtype_list(n)%face_nodes(5)%node(4))
gtype_list(n)%face_nodes(5)%node = [ 1, 2, 6, 5 ]
allocate(gtype_list(n)%face_nodes(6)%node(4))
gtype_list(n)%face_nodes(6)%node = [ 3, 4, 8, 7 ]
gtype_list(n)%supported = .true.

n = 6
gtype_list(n)%name = "6-node prism"
gtype_list(n)%dimensions = 3
gtype_list(n)%nnodes = 6
gtype_list(n)%nfaces = 5
allocate(gtype_list(n)%face_nodes(gtype_list(n)%nfaces))
allocate(gtype_list(n)%face_nodes(1)%node(4))
gtype_list(n)%face_nodes(1)%node = [ 1, 2, 5, 4 ]
allocate(gtype_list(n)%face_nodes(2)%node(4))
gtype_list(n)%face_nodes(2)%node = [ 2, 3, 6, 5 ]
allocate(gtype_list(n)%face_nodes(3)%node(4))
gtype_list(n)%face_nodes(3)%node = [ 1, 4, 6, 3 ]
allocate(gtype_list(n)%face_nodes(4)%node(3))
gtype_list(n)%face_nodes(4)%node = [ 1, 3, 2 ]
allocate(gtype_list(n)%face_nodes(5)%node(3))
gtype_list(n)%face_nodes(5)%node = [ 4, 5, 6 ]
gtype_list(n)%supported = .true.

n = 7
gtype_list(n)%name = "5-node pyramid"
gtype_list(n)%dimensions = 3
gtype_list(n)%nnodes = 5
gtype_list(n)%nfaces = 5
allocate(gtype_list(n)%face_nodes(gtype_list(n)%nfaces))
allocate(gtype_list(n)%face_nodes(1)%node(3))
gtype_list(n)%face_nodes(1)%node = [ 1, 2, 5 ]
allocate(gtype_list(n)%face_nodes(2)%node(3))
gtype_list(n)%face_nodes(2)%node = [ 2, 3, 5 ]
allocate(gtype_list(n)%face_nodes(3)%node(3))
gtype_list(n)%face_nodes(3)%node = [ 3, 4, 5 ]
allocate(gtype_list(n)%face_nodes(4)%node(3))
gtype_list(n)%face_nodes(4)%node = [ 1, 5, 4 ]
allocate(gtype_list(n)%face_nodes(5)%node(4))
gtype_list(n)%face_nodes(5)%node = [ 1, 4, 3, 2 ]
gtype_list(n)%supported = .true.

! setup gmesh option lists

allocate(output_gmesh_options(5),data_options(3),input_gmesh_options(5),output_options(2),input_options(2))
! pedantic gfortran wants equal length strings here
output_gmesh_options = ["output            ", &
                        "centringoutput    ", &
                        "meshoutput        ", &
                        "centringmeshoutput", &
                        "nooutput          "]
input_gmesh_options = ["input            ", &
                       "centringinput    ", &
                       "meshinput        ", &
                       "centringmeshinput", &
                       "noinput          "]
output_options = ["output        ", &
                  "nooutput      "]
data_options = ["elementdata           ", &
                "elementnodedata       ", &
                "elementnodelimiteddata"]
input_options = ["input            ", &
                 "noinput          "]

if (debug) write(*,'(a/80(1h+))') 'subroutine setup_gtype_list'

end subroutine setup_gtype_list

!-----------------------------------------------------------------

subroutine read_gmesh(contents,var_number)

use general_module
character(len=4) :: contents ! either mesh or data
integer, optional :: var_number ! if reading data then this is the var_number to read
integer :: gmesh_number
character(len=1000) :: formatline
character(len=100) :: input_option
logical :: success, data_input
logical, parameter :: debug = .false.

if (debug) write(*,'(80(1h+)/a)') 'subroutine read_gmesh'

if (.not.allocated(gmesh)) stop 'ERROR: no gmsh meshes have been specified'

success = .false.
data_input = .false.

do gmesh_number = 0, ubound(gmesh,1)
  formatline = '(a,'//trim(dindexformat(gmesh_number))//',a)'
  input_option = check_option(gmesh(gmesh_number)%options,input_gmesh_options) ! get input option for gmesh
  if (trim(input_option) == "noinput" .or. &
    (contents == 'data'.and.(trim(input_option) == "centringmeshinput".or.trim(input_option) == "meshinput"))) then
    if (gmesh_number /= 0.or.debug) &
      write(*,fmt=formatline) 'INFO: skipping '//trim(contents)//' read for gmesh ',gmesh_number,' having basename '// &
      trim(gmesh(gmesh_number)%basename)//' based on gmesh input option '//trim(input_option)
  else if (contents == 'mesh') then
    write(*,fmt=formatline) 'INFO: performing mesh read for gmesh ',gmesh_number,' having basename '// &
      trim(gmesh(gmesh_number)%basename)//' based on gmesh input option '//trim(input_option)
    success = .true.
    if (trim(input_option) == "centringinput" .or. trim(input_option) == "centringmeshinput") then
      write(*,'(a)') 'INFO: reading in cell element mesh:'
      call read_gmesh_file(gmesh_number=gmesh_number,file_centring='cell',contents=contents) ! read in cell mesh for gmesh_number
      write(*,'(a)') 'INFO: reading in face element mesh:'
      call read_gmesh_file(gmesh_number=gmesh_number,file_centring='face',contents=contents) ! read in face mesh for gmesh_number
    else
      call read_gmesh_file(gmesh_number=gmesh_number,file_centring='both',contents=contents) ! read in combined mesh for gmesh_number
    end if
  else if (contents == 'data') then
    data_input = .true.
    if (.not.present(var_number)) call error_stop("read_gmesh called with data contents but no var_number")
    if (trim(check_option(var(var_number)%options,input_options)) == "noinput".and. &
      trim(check_option(compound(var(var_number)%compound_number)%options,input_options)) == "noinput") then
      if (debug) write(*,fmt=formatline) 'INFO: skipping data read for variable '//trim(var(var_number)%name)//' for gmesh ', &
        gmesh_number,' having basename '//trim(gmesh(gmesh_number)%basename)//' based on component option '// &
        trim(check_option(var(var_number)%options,input_options))//' and compound option '// &
        trim(check_option(compound(var(var_number)%compound_number)%options,input_options)) 
      success = .true. ! no reading of data is a success here
    else
      if (debug) write(*,fmt=formatline) 'INFO: initiating data read for variable '//trim(var(var_number)%name)//' for gmesh ', &
        gmesh_number,' having basename '//trim(gmesh(gmesh_number)%basename)//' based on gmesh input option '// &
        trim(input_option)//', component option '// &
        trim(check_option(var(var_number)%options,input_options))//' and compound option '// &
        trim(check_option(compound(var(var_number)%compound_number)%options,input_options)) 
      if (trim(input_option) == "centringinput") then
        if (debug) write(*,'(a)') 'INFO: reading in '//var(var_number)%centring//' element data:'
        call read_gmesh_file(gmesh_number=gmesh_number,file_centring=var(var_number)%centring,contents=contents, &
          var_number=var_number,success=success) ! read in centred data for gmesh_number
      else
        call read_gmesh_file(gmesh_number=gmesh_number,file_centring='both',contents=contents,var_number=var_number, &
          success=success) ! read in combined data for gmesh_number
      end if
    end if
  end if
end do

if (contents == 'mesh'.and..not.success) write(*,'(a)') "WARNING: no meshes have been read in"
if (contents == 'data'.and.data_input.and..not.success) write(*,'(a)') &
  "WARNING: although requested, no data was read in for variable "//trim(var(var_number)%name)

if (debug) write(*,'(a/80(1h-))') 'subroutine read_gmesh'

end subroutine read_gmesh

!-----------------------------------------------------------------

subroutine read_gmesh_file(gmesh_number,file_centring,contents,var_number,success)

! this subroutine reads in a gmsh file, for contents = mesh or data
! specifically for contents=mesh it:
! - defines any regions not already defined (name, centring and location)
! - defines nodes (x, jface) and sets ktotal
! - defines cells (dimensions, jface, knode, gtype), adds to relevant region list and sets itotal
! - defines faces (dimensions, icell, knode, gtype), adds to relevant region list and sets jtotal

use general_module
integer :: gmesh_number, gmsh_bin, n, error
integer, optional :: var_number
logical, optional :: success
real :: gmsh_version
integer, parameter :: fgmsh = 2
character(len=1000) :: textline, filename
character(len=4) :: file_centring, contents
logical, parameter :: debug = .false.

if (debug) write(*,'(80(1h+)/a)') 'subroutine read_gmesh_file'

if (contents == 'data'.and..not.present(success)) stop &
  "ERROR: read_gmesh_file must be called with success logical when reading data"

filename = trim(gmesh(gmesh_number)%filename)
! for centringinput alter filename to suit either cell or face
if (file_centring == "cell" .or. file_centring == "face" .or. file_centring == "none") then
  n = scan(filename,'.',.true.)
  if (n == 0) call error_stop("problem constructing centring filename in subroutine read_gmesh_file: "//trim(filename))
  filename = filename(1:n)//trim(file_centring)//"."//filename(n+1:len(filename))
end if
if (debug) write(*,*) ' reading in filename = '//trim(filename)
open(unit=fgmsh,file=trim(filename),status='old',iostat=error)
if (error /= 0) call error_stop('problem opening mesh file '//trim(filename))

!---------------------------------------------
! check version and format of file
do 
  read(fgmsh,'(a)',iostat=error) textline
  if (error /= 0) call error_stop('gmsh file version not found for '//trim(filename))
  if (trim(textline) == "$MeshFormat") exit
end do
read(fgmsh,*,iostat=error) gmsh_version, gmsh_bin
if (error /= 0) call error_stop('problem reading in gmsh file version in '//trim(filename))
if (gmsh_version < 2.1) then
  write(*,'(a,f4.1,a)') 'ERROR: the gmsh file version (',gmsh_version,') is too old and is not supported (update to v2.1) in '// &
    trim(filename)
  stop
else if (gmsh_version > 2.2) then
  if (contents == 'mesh') write(*,'(a,f4.1,a)') 'WARNING: the gmsh mesh file version (',gmsh_version, &
    ') is not fully suported yet (but may work) in '//trim(filename)
end if
if (gmsh_bin /= 0) call error_stop('gmsh binary file type not supported in '//trim(filename))
if (debug) write(*,*) 'gmsh_version = ',gmsh_version
      
if (contents == 'mesh') then
!---------------------------------------------
! read in physical names and possibly create any required regions
! if reading the face file check that each physical region is defined

  if (file_centring /= "face") then
    call read_gmesh_regions(fgmsh=fgmsh,check=.false.,gmesh_number=gmesh_number,filename=filename,debug=debug) ! read in
    if (debug) write(*,*) 'gregions checked'
  else
    call read_gmesh_regions(fgmsh=fgmsh,check=.true.,gmesh_number=gmesh_number,filename=filename,debug=debug) ! check for consistency
    if (debug) write(*,*) 'gregions read in'
  end if

!---------------------------------------------
! read in nodes
! if reading the face file check that nodes are exactly the same as for the cell file

  if (file_centring /= "face") then
    call read_gmesh_nodes(fgmsh=fgmsh,check=.false.,gmesh_number=gmesh_number,filename=filename,debug=debug) ! read in
    if (debug) write(*,*) 'nodes checked'
  else
    call read_gmesh_nodes(fgmsh=fgmsh,check=.true.,gmesh_number=gmesh_number,filename=filename,debug=debug) ! check for consistency
    if (debug) write(*,*) 'nodes read in'
  end if

!---------------------------------------------
! read in elements, creating cells and faces as we go
  call read_gmesh_elements(fgmsh=fgmsh,gmesh_number=gmesh_number,filename=filename,debug=debug)
  if (debug) write(*,*) 'elements read in'

!---------------------------------------------
else if (contents == 'data') then
! read in data
  if (.not.present(var_number)) call error_stop("read_gmesh_file called with data contents but no var_number")
  call read_gmesh_data(fgmsh=fgmsh,gmesh_number=gmesh_number,filename=filename,var_number=var_number,success=success,debug=debug)
  if (debug) write(*,*) 'data read in'

end if
!---------------------------------------------

close(fgmsh)

if (debug) write(*,'(a/80(1h-))') 'subroutine read_gmesh_file'

end subroutine read_gmesh_file

!-----------------------------------------------------------------

subroutine read_gmesh_data(fgmsh,gmesh_number,filename,var_number,success,debug)

use general_module
integer :: fgmsh, error, gmesh_number, ntags, var_number, compound_number, nrank, nelements, gelement, nelement, m, n, k, ij, ns, &
  region_number, nnodes
character(len=1000) :: textline, filename, formatline
character(len=100) :: data_type, name
character(len=4) :: centring
double precision, dimension(:,:), allocatable :: data_array
logical :: compoundl, success
logical :: debug

if (debug) write(*,'(a)') 'INFO: reading in data for variable '//trim(var(var_number)%name)//' from file '//trim(filename)//':'

rewind(fgmsh)

data_type = ''

main_loop: do 
  read(fgmsh,'(a)',iostat=error) textline
  if (error /= 0) exit
  if (data_type /= '') then
    if (trim(textline) == '$End'//trim(data_type)) data_type = ''
  else if (trim(textline) == "$Data" .or. trim(textline) == "$ElementData" .or. trim(textline) == "$ElementNodeData") then
    data_type = trim(textline(2:len(textline)))

! character data: ie, variable name
    read(fgmsh,*,iostat=error) ntags ! ntags is the number of character variables, the first of which should be the variable name
    if (error /= 0.or.ntags < 1) then
      write(*,*) 'WARNING: problem reading file for data type '//trim(data_type)//' in gmsh file '//trim(filename)
      cycle main_loop
    end if
    do n = 1, ntags
      read(fgmsh,*,iostat=error) textline
      if (error /= 0) then
        write(*,*) 'WARNING: problem reading file for data type '//trim(data_type)//' in gmsh file '//trim(filename)
        cycle main_loop
      end if
      if (n == 1) then
        name = var(var_number)%name
        centring = var(var_number)%centring
        region_number = var(var_number)%region_number
        if (trim(textline) == trim(var(var_number)%name)) then
          compoundl = .false.
          compound_number = 0
        else if (trim(textline) == trim(compound(var(var_number)%compound_number)%name)) then
          compoundl = .true.
          compound_number = var(var_number)%compound_number
        else
          cycle main_loop ! this data field does not contain the requested var so move on
        end if
      else
        write(*,*) 'WARNING: ignoring character string in '//trim(filename)//' in gmsh file'//trim(textline)
      end if
    end do
    if (debug) then
      if (compoundl) then
        write(*,*) 'found compound variable '//trim(name)//' for var '//trim(var(var_number)%name)//' in gmsh file '//trim(filename)
      else
        write(*,*) 'found var variable '//trim(name)//' in gmsh file '//trim(filename)
      end if
    end if

! perform some sanity checks on data_type
    if ((data_type == 'Data'.and.centring /= 'none').or.(data_type /= 'Data'.and.centring == 'none').or. &
      (data_type == 'ElementNodeData'.and.centring /= 'cell')) then
      write(*,*) 'WARNING: inconsistent variable centring for variable '//trim(name)//' in gmsh file '//trim(filename)
      cycle main_loop
    end if
  
! real data: empty
    read(fgmsh,*,iostat=error) ntags ! ntags is the number of real variables, should be 0
    if (error /= 0) then
      write(*,*) 'WARNING: problem reading file for variable '//trim(name)//' in gmsh file '//trim(filename)
      cycle main_loop
    end if
    do n = 1, ntags
      read(fgmsh,*,iostat=error) textline
    end do

! integer data: tstep, number of entries per elements, number of elements
    read(fgmsh,*,iostat=error) ntags ! ntags is the number of real variables, should be 3
    if (error /= 0.or.ntags < 3) then
      write(*,*) 'WARNING: problem reading file for variable '//trim(name)//' in gmsh file '//trim(filename)
      cycle main_loop
    end if
    read(fgmsh,*,iostat=error) tstep
    if (error /= 0) then
      write(*,*) 'WARNING: problem reading file for variable '//trim(name)//' in gmsh file '//trim(filename)
      cycle main_loop
    end if
    read(fgmsh,*,iostat=error) nrank
    if (error /= 0) then
      write(*,*) 'WARNING: problem reading file for variable '//trim(name)//' in gmsh file '//trim(filename)
      cycle main_loop
    end if
    read(fgmsh,*,iostat=error) nelements
    if (error /= 0) then
      write(*,*) 'WARNING: problem reading file for variable '//trim(name)//' in gmsh file '//trim(filename)
      cycle main_loop
    end if

! perform a sanity check on variable rank
    if (compoundl) then
      if (nrank /= ubound(compound(compound_number)%component,1)) then
        write(*,*) 'WARNING: rank of compound '//trim(name)//' does not match that in gmsh file '//trim(filename)
        cycle main_loop
      end if
    else
      if (nrank /= 1) then
        write(*,*) 'WARNING: rank of component variable '//trim(name)//' does not match that in gmsh file '//trim(filename)
        cycle main_loop
      end if
    end if

    if (debug) write(*,*) ' about to read in data for variable = '//trim(name)//': compoundl = ',compoundl,': var_number = ', &
      var_number,': compound_number = ',compound_number,': nrank = ',nrank,': nelements = ',nelements,': data_type = '// &
      trim(data_type)

! do a temporary allocation of data_array, assuming one element
    allocate(data_array(nrank,1))
    ns = 1

! now to read in actual data
! errors found from here onwards indicate a real problem, so terminate if any found
    do nelement = 1, nelements
      read(fgmsh,'(a)',iostat=error) textline
      if (error /= 0) call error_stop ('problem reading file for variable '//trim(name)//' in gmsh file '//trim(filename))
      if (trim(data_type) == 'ElementNodeData') then
        read(textline,*,iostat=error) gelement, ntags
      else
        read(textline,*,iostat=error) gelement
        ntags = 1
      end if
      if (error /= 0) call error_stop ('problem reading file for variable '//trim(name)//' in gmsh file '//trim(filename))
      if (gelement < 1.or.gelement > ubound(gmesh(gmesh_number)%gelement,1)) &
        call error_stop('gelement out of range for variable '//trim(name)//' in gmsh file '//trim(filename))
      if (ntags < 1) call error_stop('ntags incorrect for variable '//trim(name)//' in gmsh file '//trim(filename))

! find cell or face if it is not none centred and check that it is within variable's region
      if (centring /= 'none') then
        ij = 0
        if (centring == 'cell') then
          ij = gmesh(gmesh_number)%gelement(gelement)%icell
        else if  (centring == 'face') then
          ij = gmesh(gmesh_number)%gelement(gelement)%jface
        end if
        if (ij == 0.and.debug) write(*,*) 'ij = ',ij,': gelement = ',gelement,': textline = ',trim(textline)
        if (ij == 0) call error_stop(trim(centring)//' centred data references an invalid element for variable '//trim(name)// &
          ' in gmsh file '//trim(filename))
        ns = region(region_number)%ns(ij)
        if (ns == 0) cycle main_loop ! silently skip any data that is not within the variable's region
      end if

! find nnodes depending on data_type and for ElementNodeData reallocate data_array to match
      nnodes = 1 ! do not have to keep reallocating data_array
      if (trim(data_type) == 'ElementNodeData') then ! NB, this also implies cell centring
        nnodes = ubound(cell(ij)%knode,1)
        if (ntags /= nnodes) call error_stop('mismatch between number of nodes that element has and what file specifies for '// &
          'variable '//trim(name)//' in gmsh file '//trim(filename))
        if (ubound(data_array,2) /= nnodes) then
          deallocate(data_array)
          allocate(data_array(nrank,nnodes))
        end if
      end if

! now read in data
      if (trim(data_type) == 'ElementNodeData') then
        read(textline,*,iostat=error) gelement, ntags, ((data_array(n,k),n=1,nrank),k=1,nnodes)
      else
        read(textline,*,iostat=error) gelement, ((data_array(n,k),n=1,nrank),k=1,nnodes)
      end if
      if (error /= 0)  call error_stop('problem reading in data line for variable '//trim(name)// &
          ' in gmsh file '//trim(filename))

! finally use data to set variable value
      if (compoundl) then
        n = location_in_list(compound(compound_number)%component,var_number)
      else
        n = 1
      end if
      if (data_type == 'Data'.or.data_type == 'ElementData') then
        var(var_number)%funk(ns)%v = data_array(n,1)
      else
! have to use cell node kernel (4) to average data for elementnodedata
        var(var_number)%funk(ns)%v = 0.d0
        do k = 1, nnodes
          var(var_number)%funk(ns)%v = var(var_number)%funk(ns)%v + cell(ij)%kernel(4)%v(k)*data_array(n,k)
        end do
      end if

    end do
    deallocate(data_array)
    if (compoundl) then
      formatline = '(a,'//trim(dindexformat(nelements))//',a,'//trim(dindexformat(nrank))//',a)'
      write(*,fmt=formatline) 'INFO: read data for variable '//trim(var(var_number)%name)//' from compound variable '// &
        trim(compound(compound_number)%name)//': centring = '//trim(centring)//': data_type = '//trim(data_type)// &
        ': nelements = ',nelements,': nrank = ',nrank,': gmsh file = '//trim(filename)
    else
      formatline = '(a,'//trim(dindexformat(nelements))//',a)'
      write(*,fmt=formatline) 'INFO: read data for variable '//trim(var(var_number)%name)// &
        ' from component: centring = '//trim(centring)//': data_type = '//trim(data_type)//': nelements = ',nelements, &
        ': gmsh file = '//trim(filename)
    end if
    success = .true.

  end if
end do main_loop

end subroutine read_gmesh_data

!-----------------------------------------------------------------

subroutine read_gmesh_elements(fgmsh,gmesh_number,filename,debug)

use general_module
integer :: fgmsh, error, gmesh_number
integer :: i, j, k, n, m, jj, nelements, nelements_tenth, gelement, gtype, ntags, region_number, nnode, nface, nfaces, &
  gelement_max, change, gelement_previous
integer, dimension(:), allocatable :: tags, local_gnodes, local_knodes
integer, dimension(:), allocatable :: icell_from_gelement, jface_from_gelement
type(cell_type) :: default_cell
type(face_type) :: default_face
character(len=1000) :: textline, filename
logical :: debug

rewind(fgmsh)

allocate(default_face%icell(2)) ! allocate spaces for the two surrounding faces

do 
  read(fgmsh,'(a)',iostat=error) textline
  if (error /= 0) call error_stop('list of elements not found in gmsh file '//trim(filename))
  if (trim(textline) == "$Elements") exit
end do
read(fgmsh,*,iostat=error) nelements
if (error /= 0) call error_stop('problem reading in number of elements in gmsh file '//trim(filename))
if (debug) write(*,*) 'number of elements: nelements = ',nelements

! allocate temporary gelement lookup arrays based on estimate of required size
allocate(icell_from_gelement(nelements))
allocate(jface_from_gelement(nelements))
gelement_max = 0 ! this records the highest gelement number
icell_from_gelement = 0
jface_from_gelement = 0

nelements_tenth = max(int(dble(nelements)/10.d0),1)
if (mod(nelements,nelements_tenth) /= 0) nelements_tenth = nelements_tenth + 1

do n = 1, nelements

  if (mod(n,nelements_tenth) == 0 .or. n == nelements) write(*,'(a,i3,a)') &
    'INFO: reading in mesh:',min(int(dble(n)*100./float(nelements)),100),'%'

  read(fgmsh,'(a)',iostat=error) textline
  if (error /= 0) call error_stop('problem reading elements in gmsh file '//trim(filename))
  read(textline,*,iostat=error) gelement, gtype, ntags
  if (error /= 0) call error_stop('problem reading initial element data in gmsh file '//trim(filename))
  if (gtype > ubound(gtype_list,1) .or. .not.gtype_list(gtype)%supported) stop &
    'ERROR: element type not supported in gmsh file'

! get tags and local_nodes arrays to the correct size ready for reading
  call resize_integer_array(array=tags,new_size=ntags)
  call resize_integer_array(array=local_gnodes,new_size=gtype_list(gtype)%nnodes)
  call resize_integer_array(array=local_knodes,new_size=gtype_list(gtype)%nnodes)

! re-read line now knowing number of tags and number of nodes
  read(textline,*,iostat=error) gelement, gtype, ntags, tags, local_gnodes
  if (error /= 0) call error_stop('problem reading element tags/nodes in gmsh file '//trim(filename))
! convert local gnode list to knode list
  do k = 1, ubound(local_gnodes,1)
    local_knodes(k) = gmesh(gmesh_number)%knode_from_gnode(local_gnodes(k))
  end do

  if (ubound(icell_from_gelement,1) < gelement) then
    change = max(gelement-ubound(icell_from_gelement,1),int(nelements/10)+1)
    call resize_integer_array(keep_data=.true.,array=icell_from_gelement,change=change)
    call resize_integer_array(keep_data=.true.,array=jface_from_gelement,change=change)
  end if
  gelement_max = max(gelement_max,gelement)
  if (debug) write(81,*) 'gelement = ',gelement,': gtype = ',gtype,': name = ',trim(gtype_list(gtype)%name), &
    ': ntags = ',ntags,': tags = ',tags,': local_gnodes = ',local_gnodes,': local_knodes = ',local_knodes

!------------------
! import as a cell

  if (gtype_list(gtype)%dimensions == gmesh(gmesh_number)%dimensions) then

    nfaces = gtype_list(gtype)%nfaces
    if (nfaces == 0) call error_stop('ERROR: a cell has no faces in the gmsh file '//trim(filename))
    call reset_cell(default_cell)
    call resize_integer_array(array=default_cell%jface,new_size=nfaces)

! loop through face node lists
    do nface = 1, nfaces

      call reset_face(default_face)
      call resize_integer_array(array=default_face%knode,new_size=ubound(gtype_list(gtype)%face_nodes(nface)%node,1))

! assemble face_node list
      do nnode = 1, ubound(default_face%knode,1)
        default_face%knode(nnode) = local_knodes(gtype_list(gtype)%face_nodes(nface)%node(nnode))
      end do

      if (debug) write(81,*) '  nface = ',nface,': default_face%knode = ',default_face%knode

! find face number if this face has been previously defined
      j = jface_from_knode_list(default_face%knode)

! if face is not found then create new one
      if (j == 0) then
        jtotal = jtotal + 1
        if (ubound(face,1) < jtotal) call resize_face_array(change=int(ktotal/10)+1)
        j = jtotal
! find gtype for face that has correct dimension and number of nodes
        default_face%gtype = 0
        do m = 1, ubound(gtype_list,1)
          if (gtype_list(m)%dimensions == gtype_list(gtype)%dimensions - 1.and. &
            gtype_list(m)%nnodes == ubound(default_face%knode,1) ) then
            default_face%gtype = m
            exit
          end if
        end do
        if (default_face%gtype == 0) call error_stop('gtype not identified for generated face in gmsh file '//trim(filename))

! set node jface references for the new face
        call add_jface_to_nodes(j,default_face%knode)
          
        if (debug) write(81,*) 'creating new face: j = ',j,': knode = ',default_face%knode,': gtype = ',default_face%gtype
      else
        default_face%gtype = face(j)%gtype
        if (debug) write(81,*) 'overwriting existing face: j = ',j,': knode = ',default_face%knode
      end if

      default_face%icell = [ 0, 0 ] ! reset icell(1) later, scrap icell(2)

! set or overwrite the face data to ensure that this cell is the icell=1 one
      default_face%dimensions = gmesh(gmesh_number)%dimensions - 1
      call set_face(face_to_set=face(j),new_value=default_face)

      default_cell%jface(nface) = j

    end do

! now check that cell doesn't exist already from knowledge of nodes
    i = icell_from_knode_list(local_knodes)

    if (i == 0) then
      default_cell%dimensions = gmesh(gmesh_number)%dimensions
      default_cell%gtype = gtype
      call copy_integer_array(original=local_knodes,copy=default_cell%knode)
      itotal = itotal + 1
      if (ubound(cell,1) < itotal) call resize_cell_array(change=int(ktotal/5)+1)
      i = itotal
      call set_cell(cell_to_set=cell(i),new_value=default_cell)
! set node icell references for the new cell
      call add_icell_to_nodes(i,default_cell%knode)
      if (debug) write(81,*) 'creating new cell: i = ',i,': jface = ',default_cell%jface
    end if

! add gmesh lookup reference
    icell_from_gelement(gelement)=i

! set icell(1) so that the surrounding face normals point in the direction of the right-hand rule for each set of face nodes
    do jj = 1, ubound(cell(i)%jface,1)
      j = cell(i)%jface(jj)
      face(j)%icell(1) = i
    end do

! add cell number to region list if it isn't there already
    if (tags(1) /= 0.and.tags(1) <= ubound(gmesh(gmesh_number)%region_from_gregion,1)) then
      region_number = gmesh(gmesh_number)%region_from_gregion(tags(1))
      if (region_number /= 0) then
        if (region(region_number)%centring /= 'cell') &
          call error_stop('gmsh cell is assigned to a region which is not cell centred in '//trim(filename))
        if (debug) write(81,*) 'cell i = ',i,': is being added to region = ',trim(region(region_number)%name)
        if (location_in_list(array=region(region_number)%ij,element=i) == 0) &
          call push_integer_array(array=region(region_number)%ij,new_element=i)
      end if
    end if

!------------------
! import as a face

  else if (gtype_list(gtype)%dimensions == gmesh(gmesh_number)%dimensions-1) then

! find face number if this face has been previously defined
    j = jface_from_knode_list(local_knodes)

! if face is not found then create new one
    if (j == 0) then
      call reset_face(default_face)
      call copy_integer_array(original=local_knodes,copy=default_face%knode)
      default_face%dimensions = gmesh(gmesh_number)%dimensions - 1
      default_face%gtype = gtype
      default_face%icell = [ 0, 0 ]
      jtotal = jtotal + 1
      if (ubound(face,1) < jtotal) call resize_face_array(change=int(ktotal/5)+1)
      j = jtotal
      call set_face(face_to_set=face(j),new_value=default_face)
      if (debug) write(81,*) 'creating new face: j = ',j,': knode = ',default_face%knode
! set node jface references for the new face
      call add_jface_to_nodes(j,default_face%knode)
    end if

! add gmesh lookup reference
    jface_from_gelement(gelement)=j

! add face number to region list if it isn't there already
    if (tags(1) /= 0.and.tags(1) <= ubound(gmesh(gmesh_number)%region_from_gregion,1)) then
      region_number = gmesh(gmesh_number)%region_from_gregion(tags(1))
      if (region_number /= 0) then
        if (region(region_number)%centring /= 'face') &
          call error_stop('gmsh face is assigned to a region which is not face centred in '//trim(filename))
        if (debug) write(81,*) 'face j = ',j,': is being added to region = ',trim(region(region_number)%name)
        if (location_in_list(array=region(region_number)%ij,element=j) == 0) &
          call push_integer_array(array=region(region_number)%ij,new_element=j)
      end if
    end if

!------------------
  else
! if meshes with multiple dimensions are present in gmsh file, this is going to cause confusion as to what is a face and what is a cell
! elements of too small dimension can be safely ignored, but there is likely to be problems with elements of the next dimension up
    write(*,*) 'WARNING: element of gtype = ',gtype,' and dimensions ',gtype_list(gtype)%dimensions,' ignored in gmsh file'
  end if

end do

! minimise cell and face arrays sizes
if (ubound(cell,1) /= itotal) call resize_cell_array(new_size=itotal)
if (ubound(face,1) /= jtotal) call resize_face_array(new_size=jtotal)

! allocate real gelement lookup arrays
! first copy info already in gelement array (from previous cell mesh reading for instance) into local variables
if (allocated(gmesh(gmesh_number)%gelement)) then
  gelement_previous = ubound(gmesh(gmesh_number)%gelement,1)
  if (gelement_previous > gelement_max) then
    call resize_integer_array(keep_data=.true.,array=icell_from_gelement,new_size=gelement_previous)
    call resize_integer_array(keep_data=.true.,array=jface_from_gelement,new_size=gelement_previous)
  end if
  do gelement = 1, gelement_previous
    if (gmesh(gmesh_number)%gelement(gelement)%icell /= 0) then
      if (icell_from_gelement(gelement) == 0) then
        icell_from_gelement(gelement) = gmesh(gmesh_number)%gelement(gelement)%icell
      else if (icell_from_gelement(gelement) /= gmesh(gmesh_number)%gelement(gelement)%icell) then
        call error_stop("cell element mismatch between elements from centringinput for "//trim(filename))
      end if
    end if
    if (gmesh(gmesh_number)%gelement(gelement)%jface /= 0) then
      if (jface_from_gelement(gelement) == 0) then
        jface_from_gelement(gelement) = gmesh(gmesh_number)%gelement(gelement)%jface
      else if (jface_from_gelement(gelement) /= gmesh(gmesh_number)%gelement(gelement)%jface) then
        call error_stop("face element mismatch between elements from centringinput for "//trim(filename))
      end if
    end if
  end do
  gelement_max = max(gelement_max,gelement_previous)
  deallocate(gmesh(gmesh_number)%gelement)
end if

! now copy local arrays to global ones
allocate(gmesh(gmesh_number)%gelement(gelement_max))
do gelement = 1, gelement_max
  gmesh(gmesh_number)%gelement(gelement)%icell = icell_from_gelement(gelement)
  gmesh(gmesh_number)%gelement(gelement)%jface = jface_from_gelement(gelement)
end do
deallocate(icell_from_gelement,jface_from_gelement)

end subroutine read_gmesh_elements

!-----------------------------------------------------------------

subroutine read_gmesh_nodes(fgmsh,check,gmesh_number,filename,debug)

use general_module
integer :: fgmsh, error, k, nnodes, gnode_check, gmesh_number, gnode_max
integer, dimension(:), allocatable :: gnode
character(len=1000) :: textline, filename
double precision, dimension(totaldimensions) :: x_check
logical :: check, debug, existing

rewind(fgmsh)
do 
  read(fgmsh,'(a)',iostat=error) textline
  if (error /= 0) call error_stop('list of nodes not found in gmsh file '//trim(filename))
  if (trim(textline) == "$Nodes") exit
end do
read(fgmsh,*,iostat=error) nnodes
if (error /= 0) call error_stop('problem reading in number of nodes in gmsh file '//trim(filename))
if (debug) write(*,*) 'number of nodes: nnodes = ',nnodes
if (.not.check) then
  call resize_node_array(change=nnodes)
  allocate(gnode(nnodes))
  do k = 1, nnodes
    read(fgmsh,*,iostat=error) gnode(k),node(k+ktotal)%x
    if (error /= 0) call error_stop('problem reading nodes in gmsh file '//trim(filename))
  end do
  ! now create lookup array
  allocate(gmesh(gmesh_number)%knode_from_gnode(maxval(gnode)))
  gmesh(gmesh_number)%knode_from_gnode = 0
  do k = 1, nnodes
    gmesh(gmesh_number)%knode_from_gnode(gnode(k)) = k + ktotal
  end do
  deallocate(gnode)
  ktotal = ktotal + nnodes
else
  gnode_max = 0
  do k = 1, nnodes
    read(fgmsh,*,iostat=error) gnode_check,x_check
    if (error /= 0) call error_stop('problem reading nodes in gmsh file '//trim(filename))
    if (distance(x_check,node(gmesh(gmesh_number)%knode_from_gnode(gnode_check))%x) > 1.d-10) call error_stop( &
      "node mismatch between msh files: cell and face msh files must come from the same simulation")
    gnode_max = max(gnode_max,gnode_check)
  end do
  if (ubound(gmesh(gmesh_number)%knode_from_gnode,1) /= gnode_max) call error_stop("number of nodes mismatched between msh "// &
    "files:  cell and face msh files must come from the same simulation")
end if

end subroutine read_gmesh_nodes

!-----------------------------------------------------------------

subroutine read_gmesh_regions(fgmsh,check,gmesh_number,filename,debug)

use general_module
integer :: fgmsh, error, ngregions, n, gmesh_number
integer, dimension(:), allocatable :: gregion_dimensions, gregion_number
character(len=1000), dimension(:), allocatable :: gregion_name
character(len=1000) :: formatline, textline, filename, location
character(len=4) :: centring
logical :: check, debug, existing

rewind(fgmsh)
do 
  read(fgmsh,'(a)',iostat=error) textline
  if (error /= 0) call error_stop('list of physical names not found in gmsh file '//trim(filename))
  if (trim(textline) == "$PhysicalNames") exit
end do
read(fgmsh,*,iostat=error) ngregions
if (error /= 0) call error_stop('problem reading in number of physical names in gmsh file '//trim(filename))
if (debug) write(*,*) 'number of physical names: ngregions = ',ngregions

! run through physical entities finding dimension and gregion_numbers
allocate(gregion_number(ngregions),gregion_dimensions(ngregions),gregion_name(ngregions))
do n=1,ngregions
  read(fgmsh,*,iostat=error) gregion_dimensions(n),gregion_number(n),gregion_name(n)
  if (debug) write(*,*) gregion_dimensions(n), gregion_number(n), trim(gregion_name(n))
  if (error /= 0) call error_stop('problem reading physical name in gmsh file '//trim(filename))
end do
if (debug) write(*,*) ' gmesh dimensions = ',maxval(gregion_dimensions)
if (check) then
  if (gmesh(gmesh_number)%dimensions /= maxval(gregion_dimensions)) call error_stop( &
    "dimension mismatch between msh files: cell and face msh files must come from the same simulation")
else
  gmesh(gmesh_number)%dimensions = maxval(gregion_dimensions)
end if

if (.not.check) then
  allocate(gmesh(gmesh_number)%region_from_gregion(maxval(gregion_number))) ! create a listing of the gmsh physical regions referenced by their gmsh number
  gmesh(gmesh_number)%region_from_gregion = 0 ! default value of 0 indicates this region is not allocated to an arb region
end if
do n=1,ngregions
  formatline = '(a,'//trim(dindexformat(gregion_number(n)))//',a)'
  write(location,fmt=formatline) 'GMSH physical entity number ',gregion_number(n),' from file '//trim(filename)
  if (gregion_dimensions(n) == gmesh(gmesh_number)%dimensions) then
    centring = "cell"
  else if (gregion_dimensions(n) == gmesh(gmesh_number)%dimensions-1) then
    centring = "face"
  else
    write(*,'(a)') 'WARNING: physical name '//trim(gregion_name(n))//' data ignored in gmsh file '//trim(filename)// &
      ' as it has too few dimensions'
    cycle
  end if
  if (check) then
    if (region_number_from_name(name=gregion_name(n), &
      location=location,centring=centring,existing=existing,creatable=.true.) == 0) call error_stop("region "// &
      trim(gregion_name(n))//" is mismatched between msh files: cell and face msh files must come from the same simulation")
  else
! now create arb region from the gregion info
    gmesh(gmesh_number)%region_from_gregion(gregion_number(n)) = region_number_from_name(name=gregion_name(n), &
      location=location,centring=centring,existing=existing,creatable=.true.)
    if (existing) then
      write(*,'(a)') "NOTE: region "//trim(gregion_name(n))//" defined in the gmsh file "//trim(filename)// &
        " conflicts with a previous region definition:  region definition in current gmsh file will be ignored"
      gmesh(gmesh_number)%region_from_gregion = 0
    else if (gmesh(gmesh_number)%region_from_gregion(gregion_number(n)) == 0) then
      write(*,*) "ERROR: problem allocating region "//trim(gregion_name(n))//" from gmsh file "//trim(filename)// &
        ":- most probably a property conflicts with that of a previously set region of the same name"
      write(*,*) 'centring = '//trim(centring)
      write(*,*) 'dimensions = ',gregion_dimensions(n)
      write(*,*) 'location = '//trim(location)
      stop
    else
! if all is OK then set dimensions
      region(gmesh(gmesh_number)%region_from_gregion(gregion_number(n)))%dimensions = gregion_dimensions(n)
    end if
  end if

end do
deallocate(gregion_number,gregion_dimensions,gregion_name)

end subroutine read_gmesh_regions

!-----------------------------------------------------------------

subroutine finalise_gmesh

! here we calculate some reverse lookups and count elements

use general_module
integer :: gmesh_number, region_number, i, j, k, n, gregion, gnode, knode, gelement, iadd, jadd, &
  new_size, old_size, jj, o
integer, dimension(:), allocatable :: face_in_gregion
type(gelement_type), dimension(:), allocatable :: gelement_old
character(len=1000) :: formatline
logical, parameter :: debug = .false.
logical :: debug_sparse = .true.
                  
if (debug) debug_sparse = .true.

if (debug_sparse) write(*,'(80(1h+)/a)') 'subroutine finalise_gmesh'

do gmesh_number = 0, ubound(gmesh,1)

  if (debug) write(*,*) 'gmesh_number = ',gmesh_number

!------------------------------
! setup output gmesh
  if (gmesh_number == 0) then

! setup output gmesh indices - completely overwrite any that were there previously
! knode and gnode have the same index
    if (allocated(gmesh(0)%knode_from_gnode)) deallocate(gmesh(0)%knode_from_gnode)
    allocate(gmesh(0)%knode_from_gnode(ktotal))
    do k = 1, ktotal
      gmesh(0)%knode_from_gnode(k) = k
    end do
! gelements are arranged as: domain cells, boundary cells/faces, domain faces
    if (allocated(gmesh(0)%gelement)) deallocate(gmesh(0)%gelement)
    allocate(gmesh(0)%gelement(itotal+jdomain))
    gmesh(0)%gelement(:)%icell = 0
    gmesh(0)%gelement(:)%jface = 0
    do i = 1, idomain
      gmesh(0)%gelement(i)%icell = i
    end do
    do i = idomain+1, itotal ! the same gmsh boundary element can refer to both the boundary cell and face
      gmesh(0)%gelement(i)%icell = i
      gmesh(0)%gelement(i)%jface = cell(i)%jface(1)
    end do
    n = itotal
    do j = 1, jtotal
      if (face(j)%type == 1) then
        n = n + 1
        if (n > itotal+jdomain) stop "ERROR: problem with setting up output gmesh in setup_mesh"
        gmesh(0)%gelement(n)%jface = j
      end if
    end do

! set output gmesh dimensions
    gmesh(0)%dimensions = maxval(cell(:)%dimensions)

! include all non-SYSTEM regions in output gmesh
    if (allocated(gmesh(0)%region_from_gregion)) deallocate(gmesh(0)%region_from_gregion)
    do region_number = 1,ubound(region,1)
      if (region(region_number)%location(1:6) /= 'SYSTEM') call push_integer_array(array=gmesh(0)%region_from_gregion, &
        new_element=region_number)
    end do

  else
!------------------------------
! operations for gmsh input gmeshes

! also include any faces that were not included originally but that border cells that are within this gmesh
    allocate(face_in_gregion(jtotal))
    face_in_gregion = 0 ! indicates face is not in region
! first mark any faces that are already given gelements as 2
    do gelement = 1, ubound(gmesh(gmesh_number)%gelement,1)
      if (gmesh(gmesh_number)%gelement(gelement)%jface /= 0) face_in_gregion(gmesh(gmesh_number)%gelement(gelement)%jface) = 2
    end do
! now loop through all cells in region, identifying faces that border these cells
    old_size = ubound(gmesh(gmesh_number)%gelement,1)
    new_size = old_size
    do gelement = 1, old_size
      i = gmesh(gmesh_number)%gelement(gelement)%icell
      if (i /= 0) then
        do jj = 1, ubound(cell(i)%jface,1)
          j = cell(i)%jface(jj)
          if (face_in_gregion(j) == 0) then
            face_in_gregion(j) = 1 ! face is in region but has not been previously defined, so define it
            new_size = new_size + 1
          end if
        end do
      end if
    end do
! now expand the gmesh%gelement array to include the new faces
    if (new_size /= old_size) then
      allocate(gelement_old(old_size))
      gelement_old = gmesh(gmesh_number)%gelement
      deallocate(gmesh(gmesh_number)%gelement)
      allocate(gmesh(gmesh_number)%gelement(new_size))
      gmesh(gmesh_number)%gelement(:)%icell = 0
      gmesh(gmesh_number)%gelement(:)%jface = 0
      gmesh(gmesh_number)%gelement(1:old_size) = gelement_old ! allocate all the old data across
      deallocate(gelement_old)
      gelement = old_size
      do j = 1, jtotal
        if (face_in_gregion(j) == 1) then
          gelement = gelement + 1
          gmesh(gmesh_number)%gelement(gelement)%jface = j
        end if
      end do
      if (gelement /= new_size) stop "ERROR: problem in finalise_gmesh"
    end if
    deallocate(face_in_gregion)

  end if
!------------------------------
! operations for all gmeshes

! reverse lookup and counting of gnodes
  call resize_integer_array(array=gmesh(gmesh_number)%gnode_from_knode,new_size=ktotal)
  gmesh(gmesh_number)%gnode_from_knode = 0
  gmesh(gmesh_number)%ngnodes = 0
  do gnode = 1, ubound(gmesh(gmesh_number)%knode_from_gnode,1)
    knode = gmesh(gmesh_number)%knode_from_gnode(gnode)
    if (knode /= 0) then
      gmesh(gmesh_number)%gnode_from_knode(knode) = gnode
      gmesh(gmesh_number)%ngnodes = gmesh(gmesh_number)%ngnodes + 1
    end if
  end do

! counting of gregions
  gmesh(gmesh_number)%ngregions = 0
  do gregion = 1, ubound(gmesh(gmesh_number)%region_from_gregion,1)
    region_number = gmesh(gmesh_number)%region_from_gregion(gregion)
    if (region_number /= 0) gmesh(gmesh_number)%ngregions = gmesh(gmesh_number)%ngregions + 1
  end do

! counting of gelements and create arrays of gregions for each gelement
  gmesh(gmesh_number)%ngelements = 0
  gmesh(gmesh_number)%ngelements_face = 0
  gmesh(gmesh_number)%ngelements_cell = 0
  do gelement = 1, ubound(gmesh(gmesh_number)%gelement,1)
    if (allocated(gmesh(gmesh_number)%gelement(gelement)%gregions)) deallocate(gmesh(gmesh_number)%gelement(gelement)%gregions)
    i = gmesh(gmesh_number)%gelement(gelement)%icell
    iadd = 0
    if (i /= 0) then
      do gregion = 1, ubound(gmesh(gmesh_number)%region_from_gregion,1)
! only include gelements that have the same dimension as the gregion (otherwise gmsh pickles)
        region_number = gmesh(gmesh_number)%region_from_gregion(gregion)
        if (region_number == 0) cycle
        if (cell(i)%dimensions /= region(region_number)%dimensions) cycle
        if (location_in_list(array=cell(i)%region_list,element=region_number) /= 0) then
          iadd = iadd + 1
          call push_integer_array(gmesh(gmesh_number)%gelement(gelement)%gregions,new_element=gregion)
        end if
      end do
    end if
    j = gmesh(gmesh_number)%gelement(gelement)%jface
    jadd = 0
    if (j /= 0) then
      do gregion = 1, ubound(gmesh(gmesh_number)%region_from_gregion,1)
        region_number = gmesh(gmesh_number)%region_from_gregion(gregion)
        if (region_number == 0) cycle
        if (face(j)%dimensions /= region(region_number)%dimensions) cycle
        if (location_in_list(array=face(j)%region_list,element=region_number) /= 0) then
          jadd = jadd + 1
          call push_integer_array(gmesh(gmesh_number)%gelement(gelement)%gregions,new_element=gregion)
        end if
      end do
    end if
    if (iadd+jadd == 0) call push_integer_array(gmesh(gmesh_number)%gelement(gelement)%gregions,new_element=0)
    if (i /= 0) gmesh(gmesh_number)%ngelements_cell = gmesh(gmesh_number)%ngelements_cell + max(iadd+jadd,1)
    if (j /= 0) gmesh(gmesh_number)%ngelements_face = gmesh(gmesh_number)%ngelements_face + max(iadd+jadd,1)
    if (i /= 0 .or. j /= 0) gmesh(gmesh_number)%ngelements = gmesh(gmesh_number)%ngelements + max(iadd+jadd,1)
  end do

  if (debug) then
    write(87,*) 'GMESH: gmesh_number = ',gmesh_number,': basename = ',trim(gmesh(gmesh_number)%basename)
    write(87,*) 'ngelements = ',gmesh(gmesh_number)%ngelements
    write(87,*) 'ngelements_cell = ',gmesh(gmesh_number)%ngelements_cell
    write(87,*) 'ngelements_face = ',gmesh(gmesh_number)%ngelements_face
    do gelement = 1, ubound(gmesh(gmesh_number)%gelement,1)
      write(87,*) 'gelement = ',gelement,': icell = ',gmesh(gmesh_number)%gelement(gelement)%icell, &
      ': jface = ',gmesh(gmesh_number)%gelement(gelement)%jface, &
      ': gregions =',(' ',gmesh(gmesh_number)%gelement(gelement)%gregions(gregion), &
      gregion=1,ubound(gmesh(gmesh_number)%gelement(gelement)%gregions,1))
    end do
  end if
    
end do

if (debug_sparse) then
  write(*,'(a)') 'INFO: gmeshes:'
  do gmesh_number = 0, ubound(gmesh,1)
    formatline = '(a,'//trim(dindexformat(gmesh_number))// &
      ',a,'//trim(dindexformat(gmesh(gmesh_number)%ngnodes))// &
      ',a,'//trim(dindexformat(gmesh(gmesh_number)%ngelements))// &
      ',a,'//trim(dindexformat(gmesh(gmesh_number)%ngelements_cell))// &
      ',a,'//trim(dindexformat(gmesh(gmesh_number)%ngelements_face))// &
      ',a,'//trim(dindexformat(gmesh(gmesh_number)%ngregions))// &
      ',a'//repeat(',a',ubound(gmesh(gmesh_number)%options,1))//')'
    write(*,fmt=formatline) ' gmesh_number = ',gmesh_number,': basename = '//trim(gmesh(gmesh_number)%basename)//': ngnodes = ', &
      gmesh(gmesh_number)%ngnodes,': ngelements = ',gmesh(gmesh_number)%ngelements,': ngelements_cell = ', &
      gmesh(gmesh_number)%ngelements_cell,': ngelements_face = ',gmesh(gmesh_number)%ngelements_face,': ngregions = ', &
      gmesh(gmesh_number)%ngregions, &
      ': options (prioritised) =',(' '//trim(gmesh(gmesh_number)%options(o)),o=1,ubound(gmesh(gmesh_number)%options,1))
  end do
end if

if (debug_sparse) write(*,'(a/80(1h-))') 'subroutine finalise_gmesh'

end subroutine finalise_gmesh

!-----------------------------------------------------------------

subroutine write_gmesh_file(gmesh_number,nstep,debug_dump,centring)

! dumps the data to a file fit for printing

use general_module
integer, parameter :: foutput = 3
integer :: gmesh_number ! output data concerned with this gmesh
integer :: nstep ! step number of newton solver
character(len=*) :: centring ! whether to output only cell, face, none or all centred elements (default all)
logical :: debug_dump ! outputs components of all variables without any gradients
integer :: error, i, j, k, l, m, region_number, kk, nnodes, n, mc, nrank, ns, &
  gregion, gnode, gelement, ngelements, ij
double precision, dimension(:), allocatable :: cellvaluel
double precision, dimension(:,:), allocatable :: cellgradl
character(len=1000) :: formatline
character(len=1000) :: filename
character(len=100) :: data_option
logical, parameter :: debug = .false.
                  
if (debug) write(*,'(80(1h+)/a)') 'subroutine write_gmesh_file'

!----------------------------------------------------
! assemble filename based on options
filename = "output/"//trim(gmesh(gmesh_number)%basename)
if (debug_dump) filename = trim(filename)//'.debug'
if (centring /= 'all')  filename = trim(filename)//'.'//centring
filename = trim(filename)//'.msh'
! open file
open(foutput,file=trim(filename),status='replace',iostat=error)
if (error /= 0) call error_stop('problem opening file '//trim(filename))

! intro stuff
! TODO: look at data-size
write(foutput,'(a/a/a)') '$MeshFormat','2.2 0 8','$EndMeshFormat'

if (centring /= 'none') then ! mesh data not required for none centred file

! list regions (irrespective of centring)
  formatline = '(a/'//trim(indexformat)//')'
  write(foutput,fmt=formatline) '$PhysicalNames',gmesh(gmesh_number)%ngregions
  do gregion = 1, ubound(gmesh(gmesh_number)%region_from_gregion,1)
    region_number = gmesh(gmesh_number)%region_from_gregion(gregion)
    if (region_number /= 0) write(foutput,'(i1,a,i3,a,a,a)') region(region_number)%dimensions,' ',gregion, &
      ' "',trim(region(region_number)%name),'"'
  end do
  write(foutput,'(a)') '$EndPhysicalNames'

! list nodes (irrespective of centring)
  formatline = '(a/'//trim(indexformat)//')'
  write(foutput,fmt=formatline) '$Nodes',gmesh(gmesh_number)%ngnodes
  formatline = '('//trim(indexformat)//',3(a,'//trim(realformat)//'))'
  do gnode = 1, ubound(gmesh(gmesh_number)%knode_from_gnode,1)
    k = gmesh(gmesh_number)%knode_from_gnode(gnode)
    if (k /= 0) write(foutput,fmt=formatline) gnode,' ',node(k)%x(1),' ',node(k)%x(2),' ',node(k)%x(3)
  end do
  write(foutput,'(a)') '$EndNodes'

! list domain cells and boundary faces as elements, using repetition to output multiple gregions
  ngelements = gmesh(gmesh_number)%ngelements
  if (centring == 'cell') ngelements = gmesh(gmesh_number)%ngelements_cell
  if (centring == 'face') ngelements = gmesh(gmesh_number)%ngelements_face
  formatline = '(a/'//trim(indexformat)//')'
  write(foutput,fmt=formatline) '$Elements', ngelements
  do gelement = 1, ubound(gmesh(gmesh_number)%gelement,1)
    i = gmesh(gmesh_number)%gelement(gelement)%icell 
    j = gmesh(gmesh_number)%gelement(gelement)%jface
    if (i /= 0.and.(centring == 'cell'.or.centring == 'all')) then
      formatline = '('//trim(indexformat)//',5(a,i3,)'//repeat(',a,'//trim(indexformat),ubound(cell(i)%knode,1))//')'
      do gregion = 1, ubound(gmesh(gmesh_number)%gelement(gelement)%gregions,1)
        write(foutput,fmt=formatline) gelement,' ',cell(i)%gtype,' ',3,' ', &
          gmesh(gmesh_number)%gelement(gelement)%gregions(gregion),' ', &
          gmesh(gmesh_number)%gelement(gelement)%gregions(gregion),' ',0, & ! set elementary entity equal to 
          (' ',gmesh(gmesh_number)%gnode_from_knode(cell(i)%knode(l)),l=1,ubound(cell(i)%knode,1))
      end do
    else if (j /= 0.and.(centring == 'face'.or.centring == 'all')) then
      formatline = '('//trim(indexformat)//',5(a,i3,)'//repeat(',a,'//trim(indexformat),ubound(face(j)%knode,1))//')'
      do gregion = 1, ubound(gmesh(gmesh_number)%gelement(gelement)%gregions,1)
        write(foutput,fmt=formatline) gelement,' ',face(j)%gtype,' ',3,' ', &
          gmesh(gmesh_number)%gelement(gelement)%gregions(gregion),' ', &
          gmesh(gmesh_number)%gelement(gelement)%gregions(gregion),' ',0, & ! set elementary entity equal to 
          (' ',gmesh(gmesh_number)%gnode_from_knode(face(j)%knode(l)),l=1,ubound(face(j)%knode,1))
      end do
    end if
  end do
  write(foutput,'(a)') '$EndElements'

end if

! if we are not output data then return here
if ((trim(check_option(gmesh(gmesh_number)%options,output_gmesh_options)) == "centringmeshoutput".and. &
  trim(check_option(gmesh(gmesh_number)%options,output_gmesh_options)) == "meshoutput").and..not.debug_dump) then
  close(foutput)
  return
end if

!---------------------------------------------------------------------------------
! component variables

do m = 1, ubound(var,1) ! loop through all component variables

  if (.not.(var(m)%centring == centring.or.centring == 'all')) cycle ! cycle if not the correct centring
  if (trim(check_option(var(m)%options,output_options)) == "nooutput".and..not.debug_dump) cycle ! cycle if we are not outputing this variable

  if (var(m)%centring == 'none') then
! none centred var components, done as Data regardless of any options

    formatline = '(a/i1/a/i1/i1/i1/i1/'//trim(indexformat)//')'
    write(foutput,fmt=formatline) '$Data',1,'"'//trim(var(m)%name)//'"',0,3,0,1,1
    formatline = '('//trim(indexformat)//',a,'//trim(realformat)//')'
    write(foutput,fmt=formatline) 1,' ',real(var(m)%funk(1)%v)
    write(foutput,'(a)') '$EndData'
    if (debug) write(91,*) 'component data (none): m = ',m,': var = ',trim(var(m)%name)

  else
! either cell or face centred

! count elements of the var's centring
    ngelements = 0
    do gelement = 1, ubound(gmesh(gmesh_number)%gelement,1)
      if (var(m)%centring == 'cell') then
        ij = gmesh(gmesh_number)%gelement(gelement)%icell
      else
        ij = gmesh(gmesh_number)%gelement(gelement)%jface
      end if
      if (ij == 0) cycle
      ns = region(var(m)%region_number)%ns(ij)
      if (ns == 0) cycle
      ngelements = ngelements + 1
    end do

! find data option - if it is specified for the gmesh, then this overwrites variable value
    data_option = trim(check_option(gmesh(gmesh_number)%options,data_options))
    if (data_option == '') data_option = trim(check_option(var(m)%options,data_options))

    if (debug_dump.or.trim(data_option) == "elementdata".or.var(m)%centring == 'face') then
! NB: face centred data can't be output with elementnodedata as facegrad is only defined in terms of cell centred data

! write out in ElementData format
      formatline = '(a/i1/a/i1/i1/i1/i1/'//trim(indexformat)//')'
      write(foutput,fmt=formatline) '$ElementData',1,'"'//trim(var(m)%name)//'"',0,3,0,1,ngelements
      do gelement = 1, ubound(gmesh(gmesh_number)%gelement,1)
        if (var(m)%centring == 'cell') then
          ij = gmesh(gmesh_number)%gelement(gelement)%icell
        else
          ij = gmesh(gmesh_number)%gelement(gelement)%jface
        end if
        if (ij == 0) cycle
        ns = region(var(m)%region_number)%ns(ij)
        if (ns == 0) cycle
        if (debug) write(91,*) 'component elementdata: gelement = ',gelement,': centring = ',var(m)%centring, &
          ': ij = ',ij,': ns = ',ns,': m = ',m,': var = ',trim(var(m)%name)
        formatline = '('//trim(indexformat)//',a,'//trim(realformat)//')'
        write(foutput,fmt=formatline) gelement,' ',real(var(m)%funk(ns)%v)
      end do
      write(foutput,'(a)') '$EndElementData'

    else if (trim(data_option) == "elementnodedata".or.trim(data_option) == "elementnodelimiteddata") then

      if (trim(data_option) == "elementnodelimiteddata") &
        write(*,*) 'WARNING: elementnodelimiteddata not implemented yet in write_gmesh'

! write out in ElementNodeData format - only for cell centred
      formatline = '(a/i1/a/i1/i1/i1/i1/'//trim(indexformat)//')'
      write(foutput,fmt=formatline) '$ElementNodeData',1,'"'//trim(var(m)%name)//'"',0,3,0,1,ngelements
      allocate(cellgradl(totaldimensions,1))
      do gelement = 1, ubound(gmesh(gmesh_number)%gelement,1)
        i = gmesh(gmesh_number)%gelement(gelement)%icell
        if (i == 0) cycle
        ns = region(var(m)%region_number)%ns(i)
        if (ns == 0) cycle
        cellgradl(:,1) = varcgrad(i,m)
        if (debug) write(91,*) 'component elementnodedata: gelement = ',gelement,': centring = ',var(m)%centring, &
          ': i = ',i,': ns = ',ns,': m = ',m,': var = ',trim(var(m)%name),': cellgrad = ',cellgradl(:,1)
        nnodes = gtype_list(cell(i)%gtype)%nnodes
        formatline = '('//trim(indexformat)//',a,i3'//repeat(',a,'//trim(realformat),nnodes)//')'
        write(foutput,fmt=formatline) gelement,' ',nnodes, &
          (' ',real(var(m)%funk(ns)%v+dot_product(cellgradl(:,1),node(cell(i)%knode(kk))%x-cell(i)%x)),kk=1,nnodes)
      end do
      deallocate(cellgradl)
      write(foutput,'(a)') '$EndElementNodeData'

    end if

  end if

end do

! return here if debug_dump only
if (debug_dump) then
  close(foutput)
  return
end if

!---------------------------------------------------------------------------------
! compound variables

do mc = 1, ubound(compound,1) ! loop through all compound variables

  if (.not.(compound(mc)%centring == centring.or.centring == 'all')) cycle ! cycle if not the correct centring
  if (trim(check_option(compound(mc)%options,output_options)) == "nooutput") cycle ! cycle if we are not outputing this variable

  nrank = ubound(compound(mc)%component,1)

  if (compound(mc)%centring == 'none') then
! none centred var components, done as Data

    formatline = '(a/i1/a/i1/i1/i1/i1/'//trim(indexformat)//')'
    write(foutput,fmt=formatline) '$Data',1,'"'//trim(compound(mc)%name)//'"',0,3,0,nrank,1
    if (debug) write(91,*) 'compound data (none): mc = ',mc,': compound = ',trim(compound(mc)%name)
    allocate(cellvaluel(nrank))
    do n = 1, nrank
      m = compound(mc)%component(n)
      if (m == 0) then
        cellvaluel(n) = 0.d0
      else
        cellvaluel(n) = var(m)%funk(1)%v
      end if
    end do
    formatline = '('//trim(indexformat)//repeat(',a,'//trim(realformat),nrank)//')'
    write(foutput,fmt=formatline) 1,(' ',real(cellvaluel(n)),n=1,nrank)
    deallocate(cellvaluel)
    write(foutput,'(a)') '$EndData'

  else
! either cell or face centred

! count elements of the var's centring
    ngelements = 0
    do gelement = 1, ubound(gmesh(gmesh_number)%gelement,1)
      if (compound(mc)%centring == 'cell') then
        ij = gmesh(gmesh_number)%gelement(gelement)%icell
      else
        ij = gmesh(gmesh_number)%gelement(gelement)%jface
      end if
      if (ij == 0) cycle
      ns = region(compound(mc)%region_number)%ns(ij)
      if (ns == 0) cycle
      ngelements = ngelements + 1
    end do

! find data option - if it is specified for the gmesh, then this overwrites variable value
    data_option = trim(check_option(gmesh(gmesh_number)%options,data_options))
    if (data_option == '') data_option = trim(check_option(compound(mc)%options,data_options))

    if (trim(data_option) == "elementdata".or.compound(mc)%centring == 'face') then

! write out in ElementData format
      formatline = '(a/i1/a/i1/i1/i1/i1/'//trim(indexformat)//')'
      write(foutput,fmt=formatline) '$ElementData',1,'"'//trim(compound(mc)%name)//'"',0,3,0,nrank,ngelements
      do gelement = 1, ubound(gmesh(gmesh_number)%gelement,1)
        if (compound(mc)%centring == 'cell') then
          ij = gmesh(gmesh_number)%gelement(gelement)%icell
        else
          ij = gmesh(gmesh_number)%gelement(gelement)%jface
        end if
        if (ij == 0) cycle
        ns = region(compound(mc)%region_number)%ns(ij)
        if (ns == 0) cycle
        if (debug) write(91,*) 'compound elementdata: gelement = ',gelement,': centring = ',compound(mc)%centring, &
          ': ij = ',ij,': ns = ',ns,': mc = ',mc,': compound = ',trim(compound(mc)%name)
        allocate(cellvaluel(nrank))
        do n = 1, nrank
          m = compound(mc)%component(n)
          if (m == 0) then
            cellvaluel(n) = 0.d0
          else
            cellvaluel(n) = var(m)%funk(ns)%v
          end if
        end do
        formatline = '('//trim(indexformat)//repeat(',a,'//trim(realformat),nrank)//')'
        write(foutput,fmt=formatline) gelement,(' ',real(cellvaluel(n)),n=1,nrank)
        deallocate(cellvaluel)
      end do
      write(foutput,'(a)') '$EndElementData'

    else if (trim(data_option) == "elementnodedata".or.trim(data_option) == "elementnodelimiteddata") then

      if (trim(data_option) == "elementnodelimiteddata") &
        write(*,*) 'WARNING: elementnodelimiteddata not implemented yet in write_gmesh'

! write out in ElementNodeData format - only for cell centred data
      formatline = '(a/i1/a/i1/i1/i1/i1/'//trim(indexformat)//')'
      write(foutput,fmt=formatline) '$ElementNodeData',1,'"'//trim(compound(mc)%name)//'"',0,3,0,nrank,ngelements
      allocate(cellvaluel(nrank))
      allocate(cellgradl(totaldimensions,nrank))
      do gelement = 1, ubound(gmesh(gmesh_number)%gelement,1)
        i = gmesh(gmesh_number)%gelement(gelement)%icell
        if (i == 0) cycle
        ns = region(compound(mc)%region_number)%ns(i)
        if (ns == 0) cycle
        if (debug) write(91,*) 'compound elementnodedata: gelement = ',gelement,': centring = ',compound(mc)%centring, &
          ': i = ',i,': ns = ',ns,': mc = ',mc,': compound = ',trim(compound(mc)%name)
        do n = 1, nrank
          m = compound(mc)%component(n)
          if (m == 0) then
            cellvaluel(n) = 0.d0
            cellgradl(:,n) = [ 0.d0, 0.d0, 0.d0 ]
          else
            cellvaluel(n) = var(m)%funk(ns)%v
            cellgradl(:,n) = varcgrad(i,m)
          end if
        end do
        nnodes = gtype_list(cell(i)%gtype)%nnodes
        formatline = '('//trim(indexformat)//',a,i3'//repeat(',a,'//trim(realformat),nnodes*nrank)//')'
        write(foutput,fmt=formatline) gelement,' ',nnodes, &
          (( ' ', real(cellvaluel(n) + dot_product(cellgradl(:,n),node(cell(i)%knode(kk))%x-cell(i)%x )), &
          n=1,nrank),kk=1,nnodes)
      end do
      deallocate(cellvaluel,cellgradl)
      write(foutput,'(a)') '$EndElementNodeData'

    end if

  end if

end do

close(foutput)

if (debug) write(*,'(80(1h+)/a)') 'subroutine write_gmesh_file'

end subroutine write_gmesh_file

!-----------------------------------------------------------------

subroutine push_gmesh(filename,gmesh_number)

! push new gmesh element onto gmesh array
! right now cannot handle allocated gregions_from_gelement

use general_module
character(len=*), intent(in) :: filename
integer, intent(out), optional :: gmesh_number
integer :: gmesh_number_local
type(gmesh_type), dimension(:), allocatable :: gmesh_tmp
integer :: n

if (allocated(gmesh)) then

! check if mesh is already defined
  do n = 0, ubound(gmesh,1)
    if (trim(gmesh(n)%basename) == trim(basename(filename))) then
      gmesh_number_local = n
      if (trim(gmesh(n)%filename) /= trim(filename)) then
        write(*,'(a)') 'ERROR: the same gmesh basename '//trim(gmesh(n)%basename)// &
          ' has been specified in multiple directories:'//'  this basename must be unique'
        stop
      end if
      if (present(gmesh_number)) gmesh_number = gmesh_number_local
      return
    end if
  end do

! if not already defined then increase gmesh array

  gmesh_number_local = ubound(gmesh,1)
  allocate(gmesh_tmp(0:gmesh_number_local))
  gmesh_tmp(0:gmesh_number_local) = gmesh(0:gmesh_number_local)
  deallocate(gmesh)
  gmesh_number_local = gmesh_number_local + 1
else
! output array should be the first one to be allocated, having index 0
  gmesh_number_local = 0
end if
allocate(gmesh(0:gmesh_number_local))
if (allocated(gmesh_tmp)) then
  gmesh(0:gmesh_number_local-1) = gmesh_tmp(0:gmesh_number_local-1)
  deallocate(gmesh_tmp)
end if

! set data for the new element
gmesh(gmesh_number_local)%filename = trim(filename)
gmesh(gmesh_number_local)%basename = trim(basename(filename))
! add default input and output options
if (gmesh_number_local == 0) then
  call push_character_array(array=gmesh(gmesh_number_local)%options,new_element='noinput')
  call push_character_array(array=gmesh(gmesh_number_local)%options,new_element='output')
else
  call push_character_array(array=gmesh(gmesh_number_local)%options,new_element='input')
  call push_character_array(array=gmesh(gmesh_number_local)%options,new_element='nooutput')
end if
if (present(gmesh_number)) gmesh_number = gmesh_number_local ! return gmesh_number if requested

end subroutine push_gmesh

!-----------------------------------------------------------------
!
!subroutine push_gregion(gregion,name,location,centring,dimensions,region_number,gregion_number)
!
!! push a new gregion onto an existing list
!
!type(gregion_type), dimension(:), allocatable :: gregion, gregion_tmp
!character(len=*), intent(in), optional :: name
!character(len=*), intent(in), optional :: location
!character(len=4), intent(in), optional :: centring
!integer, intent(in), optional :: dimensions
!integer, intent(in), optional :: region_number
!integer, intent(out), optional :: gregion_number
!integer :: gregion_number_local
!
!if (allocated(gregion)) then
!  gregion_number_local = ubound(gregion,1)
!  allocate(gregion_tmp(gregion_number_local))
!  gregion_tmp(gregion_number_local) = gregion(gregion_number_local)
!  deallocate(gregion)
!  gregion_number_local = gregion_number_local + 1
!else
!  gregion_number_local = 1
!end if
!allocate(gregion(gregion_number_local))
!if (allocated(gregion_tmp)) then
!  gregion(gregion_number_local-1) = gregion_tmp(gregion_number_local-1)
!  deallocate(gregion_tmp)
!end if
!
!gregion(gregion_number_local)%name = ''
!gregion(gregion_number_local)%location = ''
!gregion(gregion_number_local)%centring = ''
!gregion(gregion_number_local)%dimensions = 0
!gregion(gregion_number_local)%region_number = 0
!if (present(name)) gregion(gregion_number_local)%name = trim(name)
!if (present(location)) gregion(gregion_number_local)%location = trim(location)
!if (present(centring)) gregion(gregion_number_local)%centring = centring
!if (present(dimensions)) gregion(gregion_number_local)%dimensions = dimensions
!if (present(region_number)) gregion(gregion_number_local)%region_number = region_number
!if (present(gregion_number)) gregion_number = gregion_number_local
!
!end subroutine push_gregion
!
!!-----------------------------------------------------------------

end module gmsh_module
!-----------------------------------------------------------------
