! file src/kernel_module.f90
!
! Copyright 2009-2011 Dalton Harvie (daltonh@unimelb.edu.au)
! 
! This file is part of arb finite volume solver, referred to as `arb'.
! 
! arb is a software package designed to solve arbitrary partial
! differential equations on unstructured meshes using the finite volume
! method.  Primarily it consists of fortran source code, perl source
! code and shell scripts.  arb replies on certain third party software
! to run, most notably the computer algebra system maxima
! <http://maxima.sourceforge.net/> which is released under the GNU GPL.
! 
! The copyright of arb is held by Dalton Harvie.
! 
! arb is released under the GNU GPL.  arb is free software: you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License (version 3) as published by the Free Software Foundation.
! You should have received a copy of the GNU General Public License
! along with arb (see file license/gpl.txt after unpacking).  If not,
! see <http://www.gnu.org/licenses/>.
! 
! arb is distributed in the hope that it will be useful, but WITHOUT
! ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
! FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
! for more details.
! 
! For full details of arb's license see the license directory.
! 
! The current homepage for the arb finite volume solver project is
! <http://www.chemeng.unimelb.edu.au/people/staff/daltonh/downloads/arb>.
!
!-------------------------------------------------------------------------
module kernel_module

!-----------------------------------------------------------------
contains

!-----------------------------------------------------------------

subroutine setup_kernels

! here we calculate the face and cell centred kernels

use general_module
integer, parameter :: fwarn = 1, fdetail = 2
integer :: i, j, l, ii, jj, ij, ii2, i2, kk, k1, k2, ku, kd, ierror, n, ijk
integer :: n_domain_kernels, n_domain_elements, n_boundary_kernels, n_boundary_elements, n_elements, &
  min_location, local_separation, kernel_separation
double precision :: dx1, dx2, h_kernel, min_value
double precision, dimension(:,:), allocatable :: r, norm
double precision, dimension(:), allocatable :: kernel_error, w
integer, dimension(:), allocatable :: separation
logical :: error
character(len=10000) :: formatline, filename
logical, parameter :: large_kernels = .false., mls_kernels = .true., cell_from_face_kernels = .true., &
  face_relative_from_absolute = .true., adaptive_kernel_size = .true.
integer, parameter :: kernel_domain_separation = 2, kernel_boundary_separation = 4
double precision, parameter :: kernel_separation_multiplier = 0.25d0
double precision, parameter :: kernel_h_multiplier = 1.0d0
double precision, parameter :: kernel_mask_limiter = -0.10d0
double precision, parameter :: kernel_minimum = 1.d-10
logical, parameter :: debug = .false., kernel_details_file = .false.
logical :: debug_sparse = .true.

if (debug) debug_sparse = .true.

if (debug_sparse) write(*,'(80(1h+)/a)') 'subroutine setup_kernels'

filename = "output/kernel_warnings.txt"
open(fwarn,file=trim(filename),status='replace',iostat=ierror)
if (ierror /= 0) then
  write(*,*) 'ERROR: problem opening file ',filename
  stop
end if

!------------------------------------------
! run through cells and faces setting kernels
! cells:
! kernel(0) is the average based on surrounding face values
! kernel(1-3) is the derivative in the 1-3rd direction based on surrounding cell values
! kernel(4) is the average based on surrounding node values
! faces:
! kernel(0) is the average based on surrounding cell values
! kernel(1-3) is the derivative in the 1-3rd direction based on surrounding cell values
! kernel(4-6) is the derivative in the face normal directions based on surrounding cell values

!------------------------------------------
! setting up face kernels

if (mls_kernels) then
  if (debug_sparse) write(*,*) 'constructing face kernels using mls method'

  do j = 1, jtotal

    if (debug) write(85,*) '----------------------------'
    if (debug) write(85,*) 'FACE: j = ',j,': face type = ',face(j)%type,': face dimensions = ',face(j)%dimensions

! find kernel(0) first, and use this mask for all face kernels
! set centring and find h_kernel

    face(j)%kernel(0)%centring = 'cell'
    if (ubound(face(j)%knode,1) > 1) then
      h_kernel = 1.d+20
      do kk = 1, ubound(face(j)%knode,1)
        do kk2 = kk+1, ubound(face(j)%knode,1)
          h_kernel = min(h_kernel,distance(node(face(j)%knode(kk))%x , node(face(j)%knode(kk2))%x))
        end do
      end do
      h_kernel = h_kernel/2.d0
    else
      h_kernel = face(j)%dx
      if (face(j)%type == 1) h_kernel = h_kernel/2.d0 ! if not a boundary face need to divide this by 2
    end if
    h_kernel = kernel_h_multiplier*h_kernel
    if (debug) write(85,*) 'h_kernel = ',h_kernel

! calculate the separation from the current face to all elements
    call degrees_of_separation(j=j,i=face(j)%icell,separation=separation)

! loop through local separations, from the user specified to the maximum possible
    local_maximum_separation = maxval(separation,1)
    if (large_kernels) then
      kernel_separation = local_maximum_separation
    else if (face(j)%type == 2) then
      kernel_separation = min(kernel_boundary_separation,local_maximum_separation)
    else 
      kernel_separation = min(kernel_domain_separation,local_maximum_separation)
    end if

    if (.not.adaptive_kernel_size) local_maximum_separation = kernel_separation

    do local_separation = kernel_separation, local_maximum_separation

      if (debug) write(85,*) 'trying local_separation = ',local_separation

! make sure that first two elements are as per icell so that boundary values correctly applied
      call resize_integer_array(array=face(j)%kernel(0)%ijk,new_size=2)
      face(j)%kernel(0)%ijk(1:2) = face(j)%icell(1:2)

      do ii = 3, ubound(separation,1)
        if (separation(ii) <= local_separation) &
          call push_integer_array(array=face(j)%kernel(0)%ijk,new_element=face(j)%icell(ii))
      end do

      call resize_double_precision_array(array=face(j)%kernel(0)%v,new_size=ubound(face(j)%kernel(0)%ijk,1))
      face(j)%kernel(0)%v = 0.d0

! create r, scale it with h_kernel, and then convert to a consistent basis
      if (allocated(r)) deallocate(r)
      allocate(r(totaldimensions,ubound(face(j)%kernel(0)%ijk,1)))
      do ii = 1, ubound(face(j)%kernel(0)%ijk,1)
        r(:,ii) = cell(face(j)%kernel(0)%ijk(ii))%x - face(j)%x
      end do
      r = r/h_kernel

! create separation scaling vector w
      if (allocated(w)) deallocate(w)
      allocate(w(ubound(face(j)%kernel(0)%ijk,1)))
      do ii = 1, ubound(face(j)%kernel(0)%ijk,1)
        ii2 = location_in_list(array=face(j)%icell,element=face(j)%kernel(0)%ijk(ii))
        if (ii2 /= 0) then
          w(ii) = kernel_separation_multiplier**(separation(ii2)-1)
        else
!         w(ii) = kernel_separation_multiplier**(maximum_kernel_separation)
          stop 'ERROR: face kernel element not found in separation array'
        end if
      end do

! construct norm and also convert to same basis
      if (allocated(norm)) deallocate(norm)
      allocate(norm(totaldimensions,2*totaldimensions))
      norm = 0.d0
      norm(1,1) = 1.d0
      norm(2,2) = 1.d0
      norm(3,3) = 1.d0
      norm(:,4) = face(j)%norm(:,1)
      norm(:,5) = face(j)%norm(:,2)
      norm(:,6) = face(j)%norm(:,3)
      call construct_orthogonal_basis(r=r,norm=norm,error=error)
      if (error) stop 'ERROR: unable to construct orthogonal basis vectors for face kernel'

      error = .false.

!     if (face(j)%type /= 2) then
        call mls_kernel(rr=r,weight=w,kernel=face(j)%kernel(0)%v,error=error)
        if (error) stop 'ERROR: problem calculating averaging mls face kernel'
!     else  
!       face(j)%kernel(0)%v(2) = 1.d0
!     end if
        
      if (minval(face(j)%kernel(0)%v) >= min(kernel_mask_limiter,0.d0) ) exit

      if (debug_sparse) then
        formatline = '(a,'//trim(indexformat)//',a,i2,a,i2,a,g9.2)'
        write(86,fmt=formatline) &
          'criteria on averaging kernel not met: j = ',j,': type = ',face(j)%type, &
          ': local_separation = ',local_separation,': minval(face(j)%kernel(0)%v) = ',minval(face(j)%kernel(0)%v)
      end if

    end do

    local_separation = min(local_separation,local_maximum_separation) ! if the loop has run its course then need this

    if ((debug_sparse).and.local_separation /= kernel_separation) &
      write(86,'(a,i2,a,g9.2)') &
      'final local_separation = ',local_separation,': minval = ',minval(face(j)%kernel(0)%v)

! print some debugging info about kernels which contain overly negative kernel elements

    if (debug.or.(debug_sparse.and.minval(face(j)%kernel(0)%v) < kernel_mask_limiter)) then
      call degrees_of_separation(j=j,i=face(j)%kernel(0)%ijk,separation=separation)
      formatline = '(a,'//trim(indexformat)//',a,i1,a,g9.2,a,i2)'
      write(86,fmt=formatline) 'j = ',j,': type = ',face(j)%type,': h_kernel = ',h_kernel, &
        ': local_separation = ',local_separation 
      do ii = 1, ubound(face(j)%kernel(0)%ijk,1)
        i = face(j)%kernel(0)%ijk(ii)
        formatline = '(a,i3,a,'//trim(indexformat)//',a'//repeat(',1x,f6.2',ubound(r,1))//',3(a,g9.2),a,i2)'
        write(86,fmt=formatline) 'ii = ',ii,': i = ',i,': r =',r(:,ii), &
          ': rmag = ',vector_magnitude(r(:,ii)),': w = ',w(ii), &
          ': v = ',face(j)%kernel(0)%v(ii),': sep. = ',separation(ii)
      end do
    end if

! for boundary cells overwrite mls kernel
    if (face(j)%type == 2) then
      face(j)%kernel(0)%v = 0.d0
      face(j)%kernel(0)%v(2) = 1.d0
    end if

    if (debug) write(85,*) 'KERNEL MASK DETAILS: j = ',j,': kernel size = ',ubound(face(j)%kernel(0)%ijk,1), &
      ': kernel centring = ',face(j)%kernel(0)%centring

! now do derviative kernels 1->6

    do l = 1, 6
      error = .false.
      call copy_kernel(original=face(j)%kernel(0),copy=face(j)%kernel(l))
      face(j)%kernel(l)%v = 0.d0
      if (l >= 4 .and. face_relative_from_absolute) then
! construct these relative kernels (4->6) from the absolute ones calculated earlier (1->3)
        do ii = 1, ubound(face(j)%kernel(0)%ijk,1)
          do l2 = 1,3
            face(j)%kernel(l)%v(ii) = face(j)%kernel(l)%v(ii) + face(j)%norm(l2,l-3)*face(j)%kernel(l2)%v(ii)
          end do
        end do
      else
! construct kernels using mls method
        if (vector_magnitude(norm(:,l)) < 1.d-10) then
          if (debug) then
            write(85,'(a)') 'norm component when expressed in basis is zero: skipping mls kernel construction'
            write(85,*) 'l = ',l,': norm(:,l) = ',norm(:,l),': vector_magnitude(norm(:,l)) = ',vector_magnitude(norm(:,l)) 
          end if
          cycle
        end if
        call mls_kernel(rr=r,weight=w,norm=norm(:,l),kernel=face(j)%kernel(l)%v,error=error)
        if (error) stop 'ERROR: problem calculating derivative mls face kernel'
        face(j)%kernel(l)%v=face(j)%kernel(l)%v/h_kernel
      end if
    end do

    if (debug) then
      do l = 0, 6
        do ii = 1, ubound(face(j)%kernel(l)%ijk,1)
          write(85,*) 'kernel elements: l = ',l,': ijk = ',face(j)%kernel(l)%ijk(ii),': v = ',face(j)%kernel(l)%v(ii)
        end do
      end do
    end if

  end do

!----------------------
! uber simple masks suitable for 1D applications only
! only 0 (average) and 4 (gradient in face direction) defined
else
  if (debug_sparse) write(*,*) 'constructing uber simple face kernels'
  do j = 1, jtotal
    do l = 0, 6
      allocate(face(j)%kernel(l)%ijk(2),face(j)%kernel(l)%v(2))
      face(j)%kernel(l)%ijk = face(j)%icell(1:2)
      face(j)%kernel(l)%v = 0.d0
      if (l == 0.or.l == 4) then
        dx1 = abs(dot_product( cell(face(j)%icell(1))%x-face(j)%x , face(j)%norm(:,1) ))
        dx2 = abs(dot_product( cell(face(j)%icell(2))%x-face(j)%x , face(j)%norm(:,1) ))
        if (l == 0) then
          face(j)%kernel(0)%v(1) = dx2/(dx1+dx2)
          face(j)%kernel(0)%v(2) = dx1/(dx1+dx2)
        else if (l == 4) then
          face(j)%kernel(4)%v(1) = -1.d0/(dx1+dx2)
          face(j)%kernel(4)%v(2) = 1.d0/(dx1+dx2)
        end if
      end if
    end do
  end do
end if

!------------------------------------------
! setting up cell kernels

if (debug_sparse) write(*,*) 'setting up cell kernel masks'
do i=1,itotal
  do l = 0, 4
    if (l == 0) then
      cell(i)%kernel(l)%centring = 'face'
      call copy_integer_array(original=cell(i)%jface,copy=cell(i)%kernel(l)%ijk)
    else if (l == 4) then
      cell(i)%kernel(l)%centring = 'node'
      call copy_integer_array(original=cell(i)%knode,copy=cell(i)%kernel(l)%ijk)
    else
      cell(i)%kernel(l)%centring = 'cell'
      if (l == 1) then
! combine masks from surrounding faces
        allocate(cell(i)%kernel(l)%ijk(1))
        cell(i)%kernel(l)%ijk(1) = i
        do jj = 1, ubound(cell(i)%jface,1)
          j = cell(i)%jface(jj)
          do ii2 = 1, ubound(face(j)%kernel(0)%ijk,1)
            i2 = face(j)%kernel(0)%ijk(ii2)
            if (location_in_list(array=cell(i)%kernel(l)%ijk,element=i2) == 0) &
              call push_array(array=cell(i)%kernel(l)%ijk,new_element=i2) ! add element if it is not already on the list
          end do
        end do
      else
! masks 2 and 3 are just copies of mask 1
        call copy_integer_array(original=cell(i)%kernel(1)%ijk,copy=cell(i)%kernel(l)%ijk)
      end if
    end if
    allocate(cell(i)%kernel(l)%v(ubound(cell(i)%kernel(l)%ijk,1)))
    cell(i)%kernel(l)%v = 0.d0
  end do
end do

!---------------------------
! construct fall-back kernels for some

!------------

!------------
! cell

if (mls_kernels) then
  if (debug_sparse) write(*,*) 'constructing cell kernels using mls method'

  do i = 1, itotal

    if (debug) write(85,*) '----------------------------'
    if (debug) write(85,*) 'CELL: i = ',i,': cell type = ',cell(i)%type,': cell dimensions = ',cell(i)%dimensions

! set h_kernel for all cell kernels based on maximum spacing between nodes on cell if cell has enough dimensions
! otherwise based on distance to neighbouring cell centroid
    if (cell(i)%dimensions > 0) then
      h_kernel = 1.d+20
      do kk = 1, ubound(cell(i)%knode,1)
        do kk2 = kk+1, ubound(cell(i)%knode,1)
          h_kernel = min(h_kernel,distance(node(cell(i)%knode(kk))%x , node(cell(i)%knode(kk2))%x))
        end do
      end do
      h_kernel = h_kernel/2.d0
    else ! this must be a boundary cell in a 1d domain, so use dx from corresponding boundary face 
      h_kernel = face(cell(i)%jface(1))%dx
      if (face(cell(i)%jface(1))%type /= 2) call error_stop('problem in setup_kernels')
    end if
    h_kernel = kernel_h_multiplier*h_kernel
    if (debug) write(85,*) 'h_kernel = ',h_kernel

    do l = 0, 4
      error = .false.

      if (debug) then
        write(85,*)
        write(85,*) 'NEW KERNEL: i = ',i,': l = ',l,': kernel size = ',ubound(cell(i)%kernel(l)%ijk,1), &
          ': kernel centring = ',cell(i)%kernel(l)%centring
      end if

!---------------
! boundary cells take on value from boundary face as set above as they are coincident
! also, in 1D boundary cells are coincident with boundary nodes also so take on that value
      if ((l == 0.or.(l == 4.and.cell(i)%dimensions == 0)).and.cell(i)%type == 2) then

        cell(i)%kernel(l)%v(1) = 1.d0

!---------------
! use mls method to create new kernels
      else if (l == 0.or.l == 4.or..not.cell_from_face_kernels) then

! create r, scale it with h_kernel, and then convert to a consistent basis
        if (l == 0.or.l == 1.or.l == 4) then
          if (allocated(r)) deallocate(r)
          allocate(r(totaldimensions,ubound(cell(i)%kernel(l)%ijk,1)))
          do ii = 1, ubound(cell(i)%kernel(l)%ijk,1)
            if (cell(i)%kernel(l)%centring.eq.'cell') then
              r(:,ii) = cell(cell(i)%kernel(l)%ijk(ii))%x - cell(i)%x
            else if (cell(i)%kernel(l)%centring.eq.'face') then
              r(:,ii) = face(cell(i)%kernel(l)%ijk(ii))%x - cell(i)%x
            else if (cell(i)%kernel(l)%centring.eq.'node') then
              r(:,ii) = node(cell(i)%kernel(l)%ijk(ii))%x - cell(i)%x
            else
              write(*,*) 'ERROR: problem in setup_kernels with cell kernel centring'
            end if
          end do
          r = r/h_kernel
! construct norm if required and also convert to same basis
          if (l == 1) then ! only has to be calculated once
            if (allocated(norm)) deallocate(norm)
            allocate(norm(totaldimensions,totaldimensions))
            norm = 0.d0
            norm(1,1) = 1.d0
            norm(2,2) = 1.d0
            norm(3,3) = 1.d0
            call construct_orthogonal_basis(r=r,norm=norm,error=error)
          else
            call construct_orthogonal_basis(r=r,error=error)
          end if
          if (error) stop 'ERROR: unable to construct orthogonal basis vectors for cell kernel'
        end if

! use mls method to construct kernels
        if (l == 0.or.l == 4) then ! average
          call mls_kernel(rr=r,kernel=cell(i)%kernel(l)%v,error=error)
        else ! derivatives
          if (vector_magnitude(norm(:,l)) < 1.d-10) then
            if (debug) then
              write(85,'(a)') 'norm component when expressed in basis is zero: skipping mls kernel construction'
              write(85,*) 'l = ',l,': norm(:,l) = ',norm(:,l),': vector_magnitude(norm(:,l)) = ',vector_magnitude(norm(:,l)) 
            end if
            cycle
          end if
          call mls_kernel(rr=r,norm=norm(:,l),kernel=cell(i)%kernel(l)%v,error=error)
          cell(i)%kernel(l)%v=cell(i)%kernel(l)%v/h_kernel
        end if
        if (error) stop 'ERROR: problem calculating cell mls kernel'

!---------------
! construct derivative kernels from surrounding face derivative and cell averaging kernels
      else

        do jj = 1, ubound(cell(i)%kernel(0)%ijk,1)
          j = cell(i)%kernel(0)%ijk(jj)
          do ii2 = 1, ubound(face(j)%kernel(l)%ijk,1)
            i2 = face(j)%kernel(l)%ijk(ii2)
            n = location_in_list(array=cell(i)%kernel(l)%ijk,element=i2)
            if (n == 0) stop 'ERROR: member of face kernel not found in cell mask kernel when summing face elements'
            cell(i)%kernel(l)%v(n) = cell(i)%kernel(l)%v(n) + cell(i)%kernel(0)%v(jj)*face(j)%kernel(l)%v(ii2)
          end do
        end do

      end if
!---------------

      if (debug) then
        do ii = 1, ubound(cell(i)%kernel(l)%ijk,1)
          write(85,*) 'kernel elements: l = ',l,': ijk = ',cell(i)%kernel(l)%ijk(ii),': v = ',cell(i)%kernel(l)%v(ii)
        end do
      end if

    end do
  end do

else
!----------------------
! uber simple masks based on number of elements
! only 0 (average from surrounding faces) and 4 (average from surrounding nodes) defined
! average from face kernel (l=0) which is inverse of number of faces in kernel
! average from node kernel (l=4) which is inverse of number of nodes in kernel
  if (debug_sparse) write(*,*) 'constructing uber simple cell kernels'
  do i=1,itotal
    do jj = 1, ubound(cell(i)%kernel(0)%ijk,1)
      cell(i)%kernel(0)%v(jj) = 1.d0/dble(ubound(cell(i)%kernel(0)%ijk,1))
    end do
    do kk = 1, ubound(cell(i)%kernel(4)%ijk,1)
      cell(i)%kernel(4)%v(kk) = 1.d0/dble(ubound(cell(i)%kernel(4)%ijk,1))
    end do
  end do
end if

if (allocated(r)) deallocate(r)
if (allocated(w)) deallocate(w)
if (allocated(norm)) deallocate(norm)
if (allocated(separation)) deallocate(separation)

!------------------------------------------
! perform efficiency measures

if (.true.) then

  if (debug_sparse) write(*,*) 'removing small kernel elements'

! count and remove zeroish kernel functions

!------------
! face

  do j = 1, jtotal
    do l = lbound(face(j)%kernel,1), ubound(face(j)%kernel,1)
      n_elements = 0
      do n = 1, ubound(face(j)%kernel(l)%ijk,1)
        if (abs(face(j)%kernel(l)%v(n)) >= kernel_minimum) then
          n_elements = n_elements + 1
          face(j)%kernel(l)%v(n_elements) = face(j)%kernel(l)%v(n)
          face(j)%kernel(l)%ijk(n_elements) = face(j)%kernel(l)%ijk(n)
        end if
      end do
! resize kernel arrays (keeping data), but keep 1 zero element to avoid problems with general handling later
      call resize_integer_array(array=face(j)%kernel(l)%ijk,new_size=max(n_elements,1))
      call resize_double_precision_array(array=face(j)%kernel(l)%v,new_size=max(n_elements,1))
      if (n_elements == 0) then
        face(j)%kernel(l)%ijk(1) = j
        face(j)%kernel(l)%v(1) = 0.d0
      end if
    end do
  end do
      
!------------
! cell

  do i = 1, itotal
    do l = lbound(cell(i)%kernel,1), ubound(cell(i)%kernel,1)
      n_elements = 0
      do n = 1, ubound(cell(i)%kernel(l)%ijk,1)
        if (abs(cell(i)%kernel(l)%v(n)) >= kernel_minimum) then
          n_elements = n_elements + 1
          cell(i)%kernel(l)%v(n_elements) = cell(i)%kernel(l)%v(n)
          cell(i)%kernel(l)%ijk(n_elements) = cell(i)%kernel(l)%ijk(n)
        end if
      end do
! resize kernel arrays (keeping data), but keep 1 zero element to avoid problems with general handling later
      call resize_array(array=cell(i)%kernel(l)%ijk,new_size=max(n_elements,1))
      call resize_array(array=cell(i)%kernel(l)%v,new_size=max(n_elements,1))
      if (n_elements == 0) then
        cell(i)%kernel(l)%ijk(1) = i
        cell(i)%kernel(l)%v(1) = 0.d0
      end if
    end do
  end do

end if

!---------------------------
! print warnings of any negative zeroth order kernels

if (debug_sparse) write(*,*) 'checking kernel consistencies and printing statistics'

formatline = '(a,'//trim(indexformat)//',a,g12.6,a,'//trim(indexformat)//')'
n = 0
min_value = 0.d0
min_location = 0
do i = 1, itotal
  if (minval(cell(i)%kernel(0)%v) < 0.d0) then
    n = n + 1
    if (minval(cell(i)%kernel(0)%v) < min_value) then
      min_value = minval(cell(i)%kernel(0)%v)
      min_location = i
    end if
  end if
end do
if (n > 0) write(fwarn,fmt=formatline) &
  'WARNING: ',n,' negative cell averaging kernel elements detected: minimum kernel element = ',min_value,': at i = ',min_location

n = 0
min_value = 0.d0
min_location = 0
do j = 1, jtotal
  if (minval(face(j)%kernel(0)%v) < 0.d0) then
    n = n + 1
    if (minval(face(j)%kernel(0)%v) < min_value) then
      min_value = minval(face(j)%kernel(0)%v)
      min_location = j
    end if
  end if
end do
if (n > 0) write(fwarn,fmt=formatline) &
  'WARNING: ',n,' negative face averaging kernel elements detected: minimum kernel element = ',min_value,': at j = ',min_location

!---------------------------
! print warnings if not enough kernels have been constructed for averaging and in each direction

do i = 1, itotal
  formatline = '(a,'//trim(indexformat)//')'
  if (maxval(abs(cell(i)%kernel(0)%v)) < 1.d-10) write(fwarn,fmt=formatline) &
    'missing averaging kernel for cell: i = ',i
  formatline = '(a,'//trim(indexformat)//',a,i1,a,i1)'
  n = 0
  do l = 1, totaldimensions
    if (maxval(abs(cell(i)%kernel(l)%v)) >= 1.d-10) n = n + 1
  end do
  if (n < cell(i)%dimensions) write(fwarn,fmt=formatline) &
    'not enough derivative kernels have been constructed for cell: i = ',i, &
    ': number constructed = ',n,': cell dimensions = ',cell(i)%dimensions
end do
  
do j = 1, jtotal
  formatline = '(a,'//trim(indexformat)//')'
  if (maxval(abs(face(j)%kernel(0)%v)) < 1.d-10) write(fwarn,fmt=formatline) &
    'missing averaging kernel for face: j = ',j
  formatline = '(a,'//trim(indexformat)//',a,i1,a,i1)'
  n = 0
  do l = 1, 2*totaldimensions
    if (maxval(abs(face(j)%kernel(l)%v)) >= 1.d-10) n = n + 1
  end do
  if (n < 2*(face(j)%dimensions+1)) write(fwarn,fmt=formatline) &
    'not enough derivative kernels have been constructed for face: j = ',j, &
    ': number constructed = ',n,': face dimensions = ',face(j)%dimensions
end do
  
!---------------------------
! check and print warnings if kernels are inconsistent

! cell kernels
formatline = '(a,'//trim(indexformat)//',a,i1,a,i1,a,'//trim(floatformat)//')'
allocate(kernel_error(1:4))
do i = 1, itotal
  do l = 0, 4
    kernel_error = 0.d0
    if (maxval(abs(cell(i)%kernel(l)%v)) > 1.d-10) then
      if (l == 0) then ! averaging kernel l = 0, face centred
        kernel_error(1) = sum(cell(i)%kernel(l)%v) - 1.d0
        do n = 1, 3
          do ij = 1, ubound(cell(i)%kernel(l)%ijk,1)
            kernel_error(n+1) = kernel_error(n+1) + cell(i)%kernel(l)%v(ij)*face(cell(i)%kernel(l)%ijk(ij))%x(n)
          end do
          kernel_error(n+1) = kernel_error(n+1) - cell(i)%x(n)
        end do
      else if (l == 4) then ! averaging kernel l = 4, node centred
        kernel_error(1) = sum(cell(i)%kernel(l)%v) - 1.d0
        do n = 1, 3
          do ij = 1, ubound(cell(i)%kernel(l)%ijk,1)
            kernel_error(n+1) = kernel_error(n+1) + cell(i)%kernel(l)%v(ij)*node(cell(i)%kernel(l)%ijk(ij))%x(n)
          end do
          kernel_error(n+1) = kernel_error(n+1) - cell(i)%x(n)
        end do
      else ! derivative kernels l = 1,2,3, cell centred
        kernel_error(1) = sum(cell(i)%kernel(l)%v)
        do n = 1, 3
          do ij = 1, ubound(cell(i)%kernel(l)%ijk,1)
            kernel_error(n+1) = kernel_error(n+1) + cell(i)%kernel(l)%v(ij)*cell(cell(i)%kernel(l)%ijk(ij))%x(n)
          end do
          if (n == l) kernel_error(n+1) = kernel_error(n+1) - 1.d0
        end do
      end if
      do n = 1, 4
        if (abs(kernel_error(n)) > 1.d-8) write(fwarn,fmt=formatline) 'ERROR: for cell(',i,')%kernel(',l,') equation ',n, &
          ' is inconsistent: kernel_error(n) = ',kernel_error(n)
      end do
    end if
  end do
end do
deallocate(kernel_error)
  
! face kernels
allocate(kernel_error(1:4))
do j = 1, jtotal
  do l = 0, 6
    kernel_error = 0.d0
    if (maxval(abs(face(j)%kernel(l)%v)) > 1.d-10) then
      if (l == 0) then ! averaging kernel l = 0, cell centred
        kernel_error(1) = sum(face(j)%kernel(l)%v) - 1.d0
        do n = 1, 3
          do ij = 1, ubound(face(j)%kernel(l)%ijk,1)
            kernel_error(n+1) = kernel_error(n+1) + face(j)%kernel(l)%v(ij)*cell(face(j)%kernel(l)%ijk(ij))%x(n)
          end do
          kernel_error(n+1) = kernel_error(n+1) - face(j)%x(n)
        end do
      else ! derivative kernels l = 1,2,3,4,5,6, cell centred
        kernel_error(1) = sum(face(j)%kernel(l)%v)
        do n = 1, 3
          do ij = 1, ubound(face(j)%kernel(l)%ijk,1)
            kernel_error(n+1) = kernel_error(n+1) + face(j)%kernel(l)%v(ij)*cell(face(j)%kernel(l)%ijk(ij))%x(n)
          end do
          if (l < 4) then
            if (n == l) kernel_error(n+1) = kernel_error(n+1) - 1.d0
          else
            kernel_error(n+1) = kernel_error(n+1) - face(j)%norm(n,l-3)
          end if
        end do
      end if
      do n = 1, 4
        if (abs(kernel_error(n)) > 1.d-8) write(fwarn,fmt=formatline) 'ERROR: for face(',j,')%kernel(',l,') equation ',n, &
          ' is inconsistent: kernel_error(n) = ',kernel_error(n)
      end do
    end if
  end do
end do
deallocate(kernel_error)
  
!------------------------------------------
! calculate some kernel statistics

formatline = '(a,i1,a,'//trim(indexformat)//',a,f5.1)'

n_elements = 0
do l = 0, 6
  n_boundary_kernels = 0
  n_boundary_elements = 0
  n_domain_kernels = 0
  n_domain_elements = 0
  do j = 1, jtotal
    if (face(j)%type == 2) then
      n_boundary_kernels = n_boundary_kernels + 1
      n_boundary_elements = n_boundary_elements + ubound(face(j)%kernel(l)%ijk,1)
    else
      n_domain_kernels = n_domain_kernels + 1
      n_domain_elements = n_domain_elements + ubound(face(j)%kernel(l)%ijk,1)
    end if
    n_elements = n_elements + ubound(face(j)%kernel(l)%ijk,1)
  end do
  write(fwarn,fmt=formatline) 'INFO: for face kernel ',l, &
    ': number of face domain element kernels = ',n_domain_kernels, &
    ': average number of elements = ',dble(n_domain_elements)/dble(max(n_domain_kernels,1))
  write(fwarn,fmt=formatline) 'INFO: for face kernel ',l, &
    ': number of face boundary element kernels = ',n_boundary_kernels, &
    ': average number of elements = ',dble(n_boundary_elements)/dble(max(n_boundary_kernels,1))
end do

do l = 0, 4
  n_boundary_kernels = 0
  n_boundary_elements = 0
  n_domain_kernels = 0
  n_domain_elements = 0
  do i = 1, itotal
    if (cell(i)%type == 2) then
      n_boundary_kernels = n_boundary_kernels + 1
      n_boundary_elements = n_boundary_elements + ubound(cell(i)%kernel(l)%ijk,1)
    else
      n_domain_kernels = n_domain_kernels + 1
      n_domain_elements = n_domain_elements + ubound(cell(i)%kernel(l)%ijk,1)
    end if
    n_elements = n_elements + ubound(cell(i)%kernel(l)%ijk,1)
  end do
  write(fwarn,fmt=formatline) 'INFO: for cell kernel ',l, &
    ': number of cell domain element kernels = ',n_domain_kernels, &
    ': average number of elements = ',dble(n_domain_elements)/dble(max(n_domain_kernels,1))
  write(fwarn,fmt=formatline) 'INFO: for cell kernel ',l, &
    ': number of cell boundary element kernels = ',n_boundary_kernels, &
    ': average number of elements = ',dble(n_boundary_elements)/dble(max(n_boundary_kernels,1))
end do

!if (debug) write(*,'(a,i8)') 'INFO: total number of kernel elements = ',n_kernels
write(*,'(a,i8)') 'INFO: total number of kernel elements = ',n_elements
write(fwarn,'(a,i8)') 'INFO: total number of kernel elements = ',n_elements

close(fwarn)

!------------------------------------------
! write out kernel details if requested

if (kernel_details_file) then
  if (debug_sparse) write(*,*) 'writing kernel details to kernel_details.txt file'

  filename = "output/kernel_details.txt"
  open(fdetail,file=trim(filename),status='replace',iostat=ierror)
  if (ierror /= 0) then
    write(*,*) 'ERROR: problem opening file ',filename
    stop
  end if

  write(fdetail,'(a)') '-----------------------------------------------------------------------'
  write(fdetail,'(a)') 'CELL KERNEL DETAILS:'
  do i = 1, itotal
    do l = lbound(cell(i)%kernel,1), ubound(cell(i)%kernel,1)
      formatline = '(a,'//trim(indexformat)//',a,i1,a,a,a'//repeat(',a,'//trim(floatformat)//',a,'// &
        trim(indexformat)//',a',ubound(cell(i)%kernel(l)%ijk,1))//')'
      write(fdetail,fmt=formatline) ' i = ',i,': l = ',l,': centring = ',cell(i)%kernel(l)%centring,': kernel v(ijk) =', &
        (' ',cell(i)%kernel(l)%v(n),'(',cell(i)%kernel(l)%ijk(n),')',n=1,ubound(cell(i)%kernel(l)%ijk,1))
    end do
  end do
  write(fdetail,'(a)') '-----------------------------------------------------------------------'
  write(fdetail,'(a)') 'FACE KERNEL DETAILS:'
  do j = 1, jtotal
    do l = lbound(face(j)%kernel,1), ubound(face(j)%kernel,1)
      formatline = '(a,'//trim(indexformat)//',a,i1,a,a,a'//repeat(',a,'//trim(floatformat)//',a,'// &
        trim(indexformat)//',a',ubound(face(j)%kernel(l)%ijk,1))//')'
      write(fdetail,fmt=formatline) ' j = ',j,': l = ',l,': centring = ',face(j)%kernel(l)%centring,': kernel v(ijk) =', &
        (' ',face(j)%kernel(l)%v(n),'(',face(j)%kernel(l)%ijk(n),')',n=1,ubound(face(j)%kernel(l)%ijk,1))
    end do
  end do

  close(fdetail)
end if

!------------------------------------------

if (debug_sparse) write(*,'(a/80(1h-))') 'subroutine setup_kernels'

end subroutine setup_kernels

!-----------------------------------------------------------------

subroutine construct_orthogonal_basis(r,norm,error)

! here we find orthogonal basis vectors for the space spanned by r, and put those vectors
!  and the optional norm array in terms of this new basis
! r and norm get resized within this routine

use general_module
use lapack_module
use numerical_recipes_module
integer :: i, j, m, n, n_plus, basis_dimension, null_dimension, nnorm
double precision, dimension(:,:), allocatable, optional :: norm
double precision, dimension(:,:), allocatable :: r, a
double precision, dimension(:), allocatable :: w
double precision, dimension(:,:), allocatable :: u, v, r_basis
integer, dimension(:), allocatable :: basis_list, null_list
double precision, dimension(totaldimensions,totaldimensions) :: basis_vectors
logical :: error
double precision, parameter :: small_element = 1.d-10
logical, parameter :: numrec = .false.
logical, parameter :: debug = .false.

if (debug) write(*,'(80(1h+)/a)') 'subroutine construct_orthogonal_basis'

error = .false.
m = totaldimensions ! number of rows in r
n = ubound(r,2) ! number of vectors in r
if (ubound(r,1) /= totaldimensions) stop 'ERROR: dimensions of r incorrect in construct_orthonormal_basis'
nnorm = 0
if (present(norm)) then
  nnorm = ubound(norm,2) ! number of vectors in norm
  if (ubound(norm,1) /= m) stop 'ERROR: dimensions of norm incorrect in construct_orthonormal_basis'
end if
if (debug) then
  write(*,*) 'j:r'
  do j = 1, m
    write(*,'(i2,100(a,g9.2))') j,':',(r(j,i),' ',i=1,n)
  end do
  if (present(norm)) then
    write(*,*) 'j:norm'
    do j = 1, m
      write(*,'(i2,100(a,g9.2))') j,':',(norm(j,i),' ',i=1,nnorm)
    end do
  end if
end if

n_plus = max(m,n) ! increase n given to svd so that a minimum of m (3) basis vectors are returned
allocate(a(m,n_plus),u(m,m),w(n_plus),v(n_plus,n_plus))
a = 0.d0
a(:,1:n) = r ! copy the first n columns of r to a
! for svd dimensions should be: a(m,n_plus), u(m,m), w(n_plus), v(n_plus,n_plus)
if (numrec) then
  call numerical_recipes_singular_value_decomposition(a=a,u=u,w=w,v=v,error=error)
else
  call lapack_singular_value_decomposition(a=a,u=u,w=w,v=v,error=error)
end if
if (error) stop 'ERROR: problem performing singular value decomposition in construct_orthogonal_basis'
if (debug) then
  write(*,*) 'done svd decomposition'
  write(*,*) 'j:u'
  do j = 1, m
    write(*,'(i2,100(a,g9.2))') j,':',(u(j,i),' ',i=1,m)
  end do
  write(*,*) 'j:w:v'
  do j = 1, n_plus
    write(*,'(i2,100(a,g9.2))') j,':',w(j),':',(v(j,i),' ',i=1,n_plus)
  end do
end if

do j = 1, m
  if (abs(w(j)) >= small_element) then
    call push_array(array=basis_list,new_element=j)
! else
  else if (vector_magnitude(u(:,j)) >= small_element ) then ! null_list vectors must be nonzero
    call push_array(array=null_list,new_element=j)
  end if
end do
! either of these could be zero
basis_dimension = allocatable_size(basis_list)
null_dimension = allocatable_size(null_list)

! check that a basis has been found that spans fully three dimensionsal test
if (basis_dimension+null_dimension /= totaldimensions) then
  write(*,*) 'ERROR: singular value decomposition was not able to find orthonormal basis vectors'
  error = .true.
end if

! now assemble the basis vectors matrix
basis_vectors = 0.d0
do i = 1, basis_dimension
  basis_vectors(:,i) = u(:,basis_list(i))
end do
do i = basis_dimension+1, basis_dimension+null_dimension
  basis_vectors(:,i) = u(:,null_list(i-basis_dimension))
end do
  
if (debug) then
  write(*,*) 'basis_dimension = ',basis_dimension,': basis_list = ',basis_list
  write(*,*) 'null_dimension = ',null_dimension,': null_list = ',null_list
  write(*,*) 'j:basis_vectors'
  do j = 1, totaldimensions
    write(*,'(i2,100(a,g9.2))') j,':',(basis_vectors(j,i),' ',i=1,totaldimensions)
  end do
end if

deallocate(a,u,v,w)
if (allocated(null_list)) deallocate(null_list)
if (allocated(basis_list)) deallocate(basis_list)
if (error) return

! now to express all the r and norm vectors in terms of the new basis
allocate(r_basis(totaldimensions,n+nnorm))
r_basis(:,1:n) = r(:,1:n)
if (present(norm)) r_basis(:,n+1:n+nnorm) = norm(:,1:nnorm)
if (.true.) then
! noting that the basis_vectors matrix is orthogonal, its inverse is equal to its transpose
  r_basis = matmul(transpose(basis_vectors),r_basis)
else if (numrec) then
  call numerical_recipes_linear_solver(a=basis_vectors,b=r_basis,error=error)
else
  call lapack_linear_solver(a=basis_vectors,b=r_basis,error=error)
end if
if (error) then
  if (debug) write(*,*) 'ERROR: problem when converting vectors to new basis in construct_orthogonal_basis'
  deallocate(r_basis)
  return
end if

if (debug) then
  write(*,*) 'j:r_basis'
  do j = 1, totaldimensions
    write(*,'(i2,100(a,g9.2))') j,':',(r_basis(j,i),' ',i=1,n+nnorm)
  end do
end if

! check that basis_dimension is consistent with r_basis
do j = basis_dimension+1, totaldimensions
  do i = 1, n
    if (abs(r_basis(j,i)) >= small_element) stop 'ERROR: non-zero r_basis r component beyond basis_dimension'
  end do
  do i = n+1, n+nnorm
    if (abs(r_basis(j,i)) >= small_element) then
      if (debug) write(*,'(a,i1,a)') &
        'WARNING: normal vector ',i-n,' cannot be expressed in the r space basis vectors'
      r_basis(:,i) = 0.d0 ! zero all elements of this normal as a marker to calling routine
    end if
  end do
end do

! reshape norm and r to return
! keep number of rows of r greater than zero
deallocate(r)
allocate(r(max(basis_dimension,1),n))
r(:,:) = r_basis(1:max(basis_dimension,1),1:n)
if (present(norm)) then
  deallocate(norm)
  allocate(norm(max(basis_dimension,1),nnorm))
  norm(:,:) = r_basis(1:max(basis_dimension,1),n+1:n+nnorm)
end if
deallocate(r_basis)

if (debug) then
  write(*,*) 'j:r'
  do j = 1, max(basis_dimension,1)
    write(*,'(i2,100(a,g9.2))') j,':',(r(j,i),' ',i=1,n)
  end do
  if (present(norm)) then
    write(*,*) 'j:norm'
    do j = 1, max(basis_dimension,1)
      write(*,'(i2,100(a,g9.2))') j,':',(norm(j,i),' ',i=1,nnorm)
    end do
  end if
end if

if (debug) write(*,'(a/80(1h-))') 'subroutine construct_orthogonal_basis'

end subroutine construct_orthogonal_basis

!-----------------------------------------------------------------

double precision function mls_radial_kernel(rr,derivative_order)

! this is the kernel required for the mls kernel method
! this kernel is 1 at rr=0, 1/2 at rr=1, approaches 0 as rr approaches infinity

implicit none
integer :: derivative_order
double precision :: rr
double precision :: pi
!integer :: n = 1
!double precision :: a = 4.d0
!integer :: n = 2
!double precision :: a = 1.50d0
integer :: n = 3
double precision :: a = 1.20d0
!integer :: n = 4
!double precision :: a = 1.d0
!integer :: n = 8
!double precision :: a = 0.65d0

pi = 4.d0*atan(1.d0)
mls_radial_kernel = 0.d0

if (derivative_order == 0) then
!  mls_radial_kernel = (cos(pi*rr/(1.d0+rr))+1.d0)/2.d0
!  mls_radial_kernel = ((cos(pi*rr/(rr+1))+1)/2)**n
   mls_radial_kernel = (cos(pi*a*rr/(a*rr+1))+1)**n/2**n
else if (derivative_order == 1) then
!  mls_radial_kernel = -pi*sin(pi*rr/(1.d0+rr))/(2.d0*(1.d0+rr)**2)
!  mls_radial_kernel = -pi*n*(cos(pi*rr/(rr+1))+1)**(n-1)*sin(pi*rr/(rr+1))/(2**n*(rr+1)**2)
!  mls_radial_kernel = -pi*a*n*(cos(pi*a*rr/(a*rr+1))+1)**(n-1)*sin(pi*a*rr/(a*rr+1))/(2**n*(a*rr+1)**2)
   mls_radial_kernel = -pi*a*dble(n)*(cos(pi*a*rr/(a*rr+1))+1)**(n-1)*sin(pi*a*rr/(a*rr+1))/(2**n*(a*rr+1)**2)
end if

end function mls_radial_kernel

!-----------------------------------------------------------------

subroutine mls_kernel(rr,weight,norm,kernel,error)

! here we calculate the zeroth and first order kernels using a mls method, adapted from
!  papers of L. Cueto-Felgueroso, 2006ish

use general_module
use lapack_module
use numerical_recipes_module

double precision, dimension(:,:) :: rr ! array of surrounding points relative to kernel centre
double precision, dimension(:), optional :: norm ! normal for first order direction
double precision, dimension(:), optional :: weight ! weighting function
double precision, dimension(:) :: kernel ! kernel to calculate
double precision, dimension(:,:), allocatable :: pp, mm, mm_inv, ww
double precision, dimension(:), allocatable :: rr_centre
integer :: spatial_dimensions, m, n_x, n, i, j, basis_order, ierror
integer, dimension(:), allocatable :: indx, ipiv
double precision :: d
logical :: error
logical, parameter :: numrec = .false.
logical, parameter :: debug = .false.

if (debug) write(88,'(80(1h+)/a)') 'subroutine mls_kernel'

kernel = 0.d0

! find spatial dimensions
!spatial_dimensions = dimensions
spatial_dimensions = ubound(rr,1)
if (present(norm)) then
  if (spatial_dimensions /= ubound(norm,1)) then
    write(*,*) 'ERROR: dimensions of rr and norm do not match in subroutine mls_kernel'
    stop
  end if
end if
! number of kernel nodes
n_x = ubound(rr,2)
if (present(weight)) then
  if (n_x /= ubound(weight,1)) then
    write(*,*) 'ERROR: number of rr vectors and weight dimensions do not match in subroutine mls_kernel'
    stop
  end if
end if

! set polynomial basis dimension by sending a test vector to mls_polynomial
!m = spatial_dimensions + 1
basis_order = 1
do 
  m = ubound(mls_polynomial(rr=rr(:,1),basis_order=basis_order),1) ! try first order polynomial basis
! if (n_x >= m) exit
  if (basis_order == 2.and.n_x >= m+1) then
    exit
  else if (basis_order == 1.and.n_x >= m) then
    exit
  end if
  basis_order = basis_order - 1
  if (basis_order == 0) stop 'ERROR: basis_order zero in mls_kernel'
end do

if (debug) then
  write(88,*) 'spatial_dimensions,m,n_x,basis_order'
  write(88,*) spatial_dimensions,m,n_x,basis_order
  if (present(norm)) then
    write(88,*) 'derivative: norm = ',norm
  else
    write(88,*) 'average: no norm'
  end if
end if

if (debug) then
  write(88,*) 'j:rr'
  do j = 1, ubound(rr,1)
    write(88,'(i2,100(a,g9.2))') j,':',(rr(j,i),' ',i=1,ubound(rr,2))
  end do
end if

! allocate various
allocate(pp(m,n_x),mm(m,m),mm_inv(m,m),ww(n_x,n_x))

! form pp
do n = 1, n_x
  pp(:,n) = mls_polynomial(rr=rr(:,n),basis_order=basis_order)
end do

if (debug) then
  write(88,*) 'j:pp'
  do j = 1, ubound(pp,1)
    write(88,'(i2,100(a,g9.2))') j,':',(pp(j,i),' ',i=1,ubound(pp,2))
  end do
end if

! form ww
ww = 0.d0
do n = 1, n_x
! if (.not.present(norm)) then
!   ww(n,n) = radial_kernel(vector_magnitude(rr(:,n)),0)
    ww(n,n) = mls_radial_kernel(vector_magnitude(rr(:,n)),0)
  if (present(weight)) ww(n,n) = ww(n,n)*weight(n)
! else
!   ww(n,n) = radial_kernel(vector_magnitude(rr(:,n)),1)
!   ww(n,n) = mls_radial_kernel(vector_magnitude(rr(:,n)),1)
! end if
end do

if (debug) then
  write(88,*) 'j:ww'
  do j = 1, ubound(ww,1)
    write(88,'(i2,100(a,g9.2))') j,':',(ww(j,i),' ',i=1,ubound(ww,2))
  end do
end if

! form mm
mm = matmul(matmul(pp,ww),transpose(pp))

! invert mm using lu decompostion by inverting identity matrix column by column
! setup mm_inv as identity matrix
mm_inv = 0.d0
do n = 1, m
  mm_inv(n,n) = 1.d0
end do
if (numrec) then
  call numerical_recipes_linear_solver(a=mm,b=mm_inv,error=error,overwritea=.false.)
else
  call lapack_linear_solver(a=mm,b=mm_inv,error=error,overwritea=.false.)
end if

if (debug) then
  write(88,*) 'j:mm_inv'
  do j = 1, ubound(mm_inv,1)
    write(88,'(i2,100(a,g9.2))') j,':',(mm_inv(j,i),' ',i=1,ubound(mm_inv,2))
  end do
end if

allocate(rr_centre(spatial_dimensions))
rr_centre = 0.d0
if (.not.present(norm)) then
  kernel = matmul(mls_polynomial(rr=rr_centre,basis_order=basis_order),matmul(mm_inv,matmul(pp,ww)))
else
  kernel = matmul(mls_polynomial(rr=rr_centre,norm=norm,basis_order=basis_order),matmul(mm_inv,matmul(pp,ww)))
end if

if (debug) then
  write(88,*) 'j:kernel'
  do j = 1, ubound(kernel,1)
    write(88,'(i2,100(a,g9.2))') j,':',kernel(j)
  end do
end if

deallocate(pp,mm,mm_inv,ww,rr_centre)

if (debug) write(88,'(a/80(1h-))') 'subroutine mls_kernel'

end subroutine mls_kernel

!-----------------------------------------------------------------

function mls_polynomial(rr,norm,basis_order)

! here we calculate mls polynomial basis function evaluated at rr
! if norm is present, calculate the gradient of this in direction of norm

use general_module
double precision, dimension(:) :: rr ! single location vector of unknown dimensions
double precision, dimension(:), optional :: norm ! normal for first order direction
!integer :: basis_order = 2
integer :: basis_order ! maximum power within basis functions
double precision, dimension(:), allocatable :: mls_polynomial ! result returned is the polynomial basis array, here first order
!double precision, dimension(ubound(rr,1)+1) :: mls_polynomial ! result returned is the polynomial basis array, here first order
integer :: spatial_dimensions, polynomial_dimensions, n
logical, parameter :: debug = .false.

if (debug) write(88,'(80(1h+)/a)') 'function mls_polynomial'

spatial_dimensions = ubound(rr,1)
if (present(norm)) then
  if (ubound(norm,1) /= spatial_dimensions) then
    write(*,*) 'norm and rr spatial dimensions don''t match in mls_polynomial'
    stop
  end if
end if
if (basis_order == 2) then
  if (spatial_dimensions == 1) then
    polynomial_dimensions = 3
  else if (spatial_dimensions == 2) then
    polynomial_dimensions = 6
  else if (spatial_dimensions == 3) then
    polynomial_dimensions = 10
  else
    stop 'ERROR: spatial dimensions incorrect in mls_polynomial'
  end if
else
  polynomial_dimensions = 1 + spatial_dimensions
end if

allocate(mls_polynomial(polynomial_dimensions))

if (.not.present(norm)) then
  mls_polynomial(1) = 1.d0
  mls_polynomial(2:spatial_dimensions+1) = rr
  if (basis_order == 2) then
    do n = 1, spatial_dimensions
      mls_polynomial(n+spatial_dimensions+1) = rr(n)**2
    end do
    if (spatial_dimensions == 2) then
      mls_polynomial(6) = rr(1)*rr(2)
    else if (spatial_dimensions == 3) then
      mls_polynomial(8) = rr(1)*rr(2)
      mls_polynomial(9) = rr(1)*rr(3)
      mls_polynomial(10) = rr(2)*rr(3)
    end if
  end if
else
  mls_polynomial(1) = 0.d0
  mls_polynomial(2:spatial_dimensions+1) = norm
  if (basis_order == 2) then
    do n = 1, spatial_dimensions
      mls_polynomial(n+spatial_dimensions+1) = 2.d0*rr(n)*norm(n)
    end do
    if (spatial_dimensions == 2) then
      mls_polynomial(6) = rr(1)*norm(2)+rr(2)*norm(1)
    else if (spatial_dimensions == 3) then
      mls_polynomial(8) = rr(1)*norm(2)+rr(2)*norm(1)
      mls_polynomial(9) = rr(1)*norm(3)+rr(3)*norm(1)
      mls_polynomial(10) = rr(2)*norm(3)+rr(3)*norm(2)
    end if
  end if
end if
  
if (debug) write(88,'(a/80(1h-))') 'function mls_polynomial'

end function mls_polynomial

!-----------------------------------------------------------------

subroutine degrees_of_separation(j,i,separation)

! here we calculate the array degrees_of_separation which is defined
!  as the number of faces you have to pass through from face j to get to the corresponding
!  cell in array i (i is the kernel array of cells surrounding face j)

use general_module
integer :: j, dimensions
integer, dimension(:), allocatable :: i
integer, dimension(:), allocatable :: separation
integer :: current_sep, n, i2, ii2, n2
integer, parameter :: maximum_separation = 10
logical, parameter :: debug = .false.

if (debug) write(88,'(80(1h+)/a)') 'subroutine degrees_of_separation'

call resize_integer_array(array=separation,new_size=ubound(i,1))
separation = 0
if (ubound(i,1) < 2) stop 'ERROR: i array too small in degrees_of_separation'
separation(1:2) = 1 ! first 2 cells assumed adjacent to face

do current_sep = 2, maximum_separation
  do n = 1, ubound(separation,1)
    if (separation(n) == current_sep-1) then
      if (debug) write(*,*) 'processing cell i(n) = ',i(n),': separation = ',separation(n)
      do ii2 = 2, ubound(cell(i(n))%icell,1) ! loop through all surrounding cells
        i2 = cell(i(n))%icell(ii2)
        n2 = location_in_list(array=i,element=i2)
        if (debug) write(*,*) ' in surrounding cell i2 = ',i2,': with n2 = ',n2
        if (n2 == 0) cycle
!       if (elements_in_common(cell(i(n))%knode,cell(i2)%knode) >= cell(i(n))%dimensions) then
! separation now strictly based on number of real faces between cells
        dimensions = max(cell(i(n))%dimensions,cell(i2)%dimensions)
        if (elements_in_common(cell(i(n))%knode,cell(i2)%knode) >= dimensions) then
          if (separation(n2) == 0) separation(n2) = current_sep ! we are next to a cell that has a separation set, so separation must be current_set
          if (debug) write(*,*) ' elements_in_common = ',elements_in_common(cell(i(n))%knode,cell(i2)%knode)
          if (debug) write(*,*) ' separation(n2) = ',separation(n2)
        end if
      end do
! based on adjacent cell relationship
!     do ii2 = 2, ubound(cell(i(n))%jface,1)+1 ! loop through only cells surrounding i(n)
!       i2 = cell(i(n))%icell(ii2)
!       n2 = location_in_list(array=i,element=i2)
!       if (debug) write(*,*) ' in surrounding cell i2 = ',i2,': with n2 = ',n2
!       if (n2 /= 0) then
!         if (separation(n2) == 0) separation(n2) = current_sep
!         if (debug) write(*,*) ' separation(n2) = ',separation(n2),': should not be zero here'
!       end if
!     end do
    end if
  end do
  if (debug) then
    write(*,*) 'at end of current_sep loop'
    write(*,*) 'current_sep = ',current_sep,': minval(sep.) = ',minval(separation)
    write(*,*) 'separation = ',separation
  end if
  if (minval(separation) /= 0) exit ! all cells have a defined separation now
end do

if (minval(separation) == 0) then
  do n = 1, ubound(separation,1)
    if (separation(n) == 0) separation(n) = maximum_separation + 1
  end do
end if

if (debug) write(*,*) 'final current_sep = ',current_sep

if (debug) write(88,'(a/80(1h-))') 'subroutine degrees_of_separation'

end subroutine degrees_of_separation

!-----------------------------------------------------------------

end module kernel_module

!-----------------------------------------------------------------
