! file src/output_module.f90
!
! Copyright 2009-2011 Dalton Harvie (daltonh@unimelb.edu.au)
! 
! This file is part of arb finite volume solver, referred to as `arb'.
! 
! arb is a software package designed to solve arbitrary partial
! differential equations on unstructured meshes using the finite volume
! method.  Primarily it consists of fortran source code, perl source
! code and shell scripts.  arb replies on certain third party software
! to run, most notably the computer algebra system maxima
! <http://maxima.sourceforge.net/> which is released under the GNU GPL.
! 
! The copyright of arb is held by Dalton Harvie.
! 
! arb is released under the GNU GPL.  arb is free software: you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License (version 3) as published by the Free Software Foundation.
! You should have received a copy of the GNU General Public License
! along with arb (see file license/gpl.txt after unpacking).  If not,
! see <http://www.gnu.org/licenses/>.
! 
! arb is distributed in the hope that it will be useful, but WITHOUT
! ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
! FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
! for more details.
! 
! For full details of arb's license see the license directory.
! 
! The current homepage for the arb finite volume solver project is
! <http://www.chemeng.unimelb.edu.au/people/staff/daltonh/downloads/arb>.
!
!-------------------------------------------------------------------------
module output_module

!-----------------------------------------------------------------
contains

!-----------------------------------------------------------------

subroutine output(nstep)

! dumps the data to a file fit for printing

use general_module
use equations_module
use gmsh_module
integer :: nstep, gmesh_number
character(len=100) :: output_option
character(len=1000) :: formatline
logical, parameter :: gmesh_output_by_centring = .false. ! true to output separate files for each centring, false for the one file
logical, parameter :: debug = .false.
                  
if (debug) write(*,'(80(1h+)/a)') 'subroutine output'

call update_outputs ! update any output-only variables

if (.false.) call output_txt(nstep) ! write txt files containing data

formatline = '(a,'//trim(dindexformat(gmesh_number))//',a)'
do gmesh_number = 0, ubound(gmesh,1)
  output_option = trim(check_option(gmesh(gmesh_number)%options,output_gmesh_options))
  if (trim(output_option) == "centringoutput" .or. trim(output_option) == "centringmeshoutput") then
    if (debug) write(*,fmt=formatline) 'INFO: writing gmesh ',gmesh_number,' having basename '//trim(gmesh(gmesh_number)%basename)
    call write_gmesh_file(gmesh_number=gmesh_number,nstep=nstep,debug_dump=.false.,centring='cell') ! write gmsh output file for cell centred variables
    call write_gmesh_file(gmesh_number=gmesh_number,nstep=nstep,debug_dump=.false.,centring='face') ! write gmsh output file for face centred variables
    call write_gmesh_file(gmesh_number=gmesh_number,nstep=nstep,debug_dump=.false.,centring='none') ! write gmsh output file for none centred variables
  else if (trim(output_option) == "output" .or. trim(output_option) == "meshoutput") then
    if (debug) write(*,fmt=formatline) 'INFO: writing gmesh ',gmesh_number,' having basename '//trim(gmesh(gmesh_number)%basename)
    call write_gmesh_file(gmesh_number=gmesh_number,nstep=nstep,debug_dump=.false.,centring='all') ! write gmsh output file for all variables
  else if (trim(output_option) == "nooutput") then
    if (debug) write(*,fmt=formatline) 'INFO: skipping write for gmesh ',gmesh_number,' having basename '//trim(gmesh(gmesh_number)%basename)
  end if
end do

! write gmsh file with all data output as components and elementdata
if (.false.) call write_gmesh_file(gmesh_number=0,nstep=nstep,debug_dump=.true.,centring='all')

if (.true.) call output_stat(nstep) ! write some statistics about the data

if (debug) write(*,'(a/80(1h-))') 'subroutine output'

end subroutine output

!-----------------------------------------------------------------

subroutine output_txt(nstep)

! dumps the data to a file fit for printing

use general_module

integer, parameter :: foutput = 3
integer :: error, i, inext, ii, n, nn, m, nstep, j, jj, l, list_length
integer, dimension(:), allocatable :: local_list
character(len=100) :: filename, cellxname(3), facexname(3)
character(len=1000) :: formatline
character(len=1) :: delimiter=' '
logical :: therel
logical, parameter :: debug = .false.
                  
if (debug) write(*,'(80(1h+)/a)') 'subroutine output_txt'

do l = 1,3
  write(cellxname(l),'(a,i1,a)') '"<cellx[l=',l,']> [?]"'
  write(facexname(l),'(a,i1,a)') '"<facex[l=',l,']> [?]"'
end do

!----------------------------------------------------
! open output file
filename = "output/output.txt"
therel = .false.
inquire (file=trim(filename),exist=therel)
open(foutput,file=trim(filename),access='append',iostat=error)
if (error /= 0) then
  write(*,*) 'ERROR: problem opening file ',filename
  stop
end if

if (therel) write(foutput,'(//)')
formatline = '(a/a,'//trim(indexformat)//'/)'
write(foutput,fmt=formatline) '#########################################','# NSTEP = ',nstep

!----------------------------------------------------
! output of cell centred general variables
list_length = allocatable_size(var_list(var_list_number(centring="cell",type="all"))%list)

if (list_length > 0) then
  allocate(local_list(list_length))
  local_list = var_list(var_list_number(centring="cell",type="all"))%list

  write(foutput,'(/a)') '# CELL CENTRED GENERAL VARIABLES'
  formatline = '(a8'//repeat(',a,'//trim(stringformat),totaldimensions+ubound(local_list,1))//')'
  write(foutput,fmt=formatline) '"icell"',(delimiter,cellxname(l),l=1,totaldimensions), &
    (delimiter,'"'//trim(var(local_list(m))%name)//' ['//trim(var(local_list(m))%units)//']"',m=1,ubound(local_list,1))
  formatline = '('//trim(indexformat)//repeat(',a,'//trim(floatformat),totaldimensions)// &
    repeat(',a,'//trim(floatformat),ubound(local_list,1))//')'
  do i = 1, itotal
    write(foutput,fmt=formatline) i,(delimiter,cell(i)%x(l),l=1,totaldimensions), &
      (delimiter,var(local_list(m))%funk(i)%v,m=1,ubound(local_list,1))
  end do

  deallocate(local_list)
end if

!----------------------------------------------------
! output of face centred general variables
list_length = allocatable_size(var_list(var_list_number(centring="face",type="all"))%list)

if (list_length > 0) then
  allocate(local_list(list_length))
  local_list = var_list(var_list_number(centring="face",type="all"))%list

  write(foutput,'(/a)') '# FACE CENTRED GENERAL VARIABLES'
  formatline = '(a8'//repeat(',a,'//trim(stringformat),totaldimensions+ubound(local_list,1))//')'
  write(foutput,fmt=formatline) '"jface"',(delimiter,facexname(l),l=1,totaldimensions), &
    (delimiter,'"'//trim(var(local_list(m))%name)//' ['//trim(var(local_list(m))%units)//']"',m=1,ubound(local_list,1))
  formatline = '('//trim(indexformat)//repeat(',a,'//trim(floatformat),totaldimensions)// &
    repeat(',a,'//trim(floatformat),ubound(local_list,1))//')'
  do j = 1, jtotal
    write(foutput,fmt=formatline) j,(delimiter,face(j)%x(l),l=1,totaldimensions), &
      (delimiter,var(local_list(m))%funk(j)%v,m=1,ubound(local_list,1))
  end do

  deallocate(local_list)
end if
!----------------------------------------------------
! output of no location general variables
list_length = allocatable_size(var_list(var_list_number(centring="none",type="all"))%list)

if (list_length > 0) then
  allocate(local_list(list_length))
  local_list = var_list(var_list_number(centring="none",type="all"))%list

  write(foutput,'(/a)') '# NONE CENTRED GENERAL VARIABLES'
  formatline = '(a8'//repeat(',a,'//trim(stringformat),ubound(local_list,1))//')'
  write(foutput,fmt=formatline) '"nstep"', &
    (delimiter,'"'//trim(var(local_list(m))%name)//' ['//trim(var(local_list(m))%units)//']"',m=1,ubound(local_list,1))

  formatline = '('//trim(indexformat)//repeat(',a,'//trim(floatformat),ubound(local_list,1))//')'
  write(foutput,fmt=formatline) nstep,(delimiter,var(local_list(m))%funk(1)%v,m=1,ubound(local_list,1))

  deallocate(local_list)
end if

close(foutput)

!----------------------------------------------------

if (debug) write(*,'(a/80(1h-))') 'subroutine output_txt'

end subroutine output_txt

!-----------------------------------------------------------------

subroutine output_stat(nstep)

! prints out some current statistics about the data

use general_module

integer, parameter :: foutput = 3
integer :: error, nstep, ns, m, max_loc, min_loc
character(len=100) :: filename
character(len=1000) :: formatline
logical :: therel
logical, parameter :: debug = .false.
                  
if (debug) write(*,'(80(1h+)/a)') 'subroutine output_stat'

filename = "output/output.stat"
therel = .false.
inquire (file=trim(filename),exist=therel)
open(foutput,file=trim(filename),access='append',iostat=error)
if (error /= 0) then
  write(*,*) 'ERROR: problem opening file ',filename
  stop
end if

if (therel) write(foutput,'(//)')
formatline = '(a/a,'//trim(indexformat)//'/)'
write(foutput,fmt=formatline) '#########################################','# NSTEP = ',nstep
  
! general variables
formatline = '(a,a,'//trim(floatformat)//',a,a,a,'//trim(indexformat)//',a,'//trim(floatformat)//',a,a,a,'//trim(indexformat)//')'
do m = 1, ubound(var,1)
  max_loc = 1
  min_loc = 1
  do ns = 2, ubound(var(m)%funk,1)
    if (var(m)%funk(ns)%v > var(m)%funk(max_loc)%v) max_loc = ns
    if (var(m)%funk(ns)%v < var(m)%funk(min_loc)%v) min_loc = ns
  end do
  write(foutput,fmt=formatline) trim(var(m)%type)//' '//trim(var(m)%name),': max ',var(m)%funk(max_loc)%v,' at ',var(m)%centring, &
    ' ',ijvar(m,max_loc),': min ',var(m)%funk(min_loc)%v,' at ',var(m)%centring,' ',ijvar(m,min_loc)
end do

close(foutput)

!----------------------------------------------------

if (debug) write(*,'(a/80(1h-))') 'subroutine output_stat'

end subroutine output_stat

!-----------------------------------------------------------------

end module output_module
